#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Student Individualized Education Plan (IEP)"""

from persistent import Persistent
from persistent.dict import PersistentDict

from zope.annotation.interfaces import IAnnotations
from zope.component import adapter, getUtility
from zope.intid.interfaces import IIntIds
from zope.interface import implementer

from schooltool.basicperson.interfaces import IBasicPerson

from schooltool.cando.interfaces import ISectionSkills
from schooltool.cando.interfaces import IStudentIEP


STUDENT_IEP_KEY = 'schooltool.cando.iep'


class StudentIEP(Persistent):

    active = False
    description = None

    def __init__(self):
        super(StudentIEP, self).__init__()
        self.iep_skills = PersistentDict({})

    def getIEPSkills(self, section):
        result = {}
        int_ids = getUtility(IIntIds)
        section_id = int_ids.getId(section)
        section_iep_skills = self.iep_skills.get(section_id)
        if section_iep_skills is not None:
            section_skills = ISectionSkills(section)
            for skillset_id, skill_ids in section_iep_skills.items():
                skillset = section_skills.get(skillset_id)
                if skillset is not None:
                    result[skillset] = []
                    for skill_id in skill_ids:
                        skill = skillset.get(skill_id)
                        if skill is not None:
                            result[skillset].append(skill)
        return result

    def addSkill(self, section, skill):
        int_ids = getUtility(IIntIds)
        section_id = int_ids.getId(section)
        section_skills = ISectionSkills(section)
        skillset = skill.__parent__
        skillset_id = skillset.__name__
        if skillset_id in section_skills:            
            if section_id not in self.iep_skills:
                self.iep_skills[section_id] = {}
            section_iep_skills = self.iep_skills.get(section_id)
            if skillset_id not in section_iep_skills:
                section_iep_skills[skillset_id] = set()
            section_iep_skills[skillset_id].add(skill.__name__)

    def removeSkill(self, section, skill):
        int_ids = getUtility(IIntIds)
        section_id = int_ids.getId(section)
        section_iep_skills = self.iep_skills.get(section_id)
        if section_iep_skills is not None:
            skill_id = skill.__name__
            skillset = skill.__parent__
            skillset_id = skillset.__name__
            if skillset_id in section_iep_skills and \
               skill_id in section_iep_skills[skillset_id]:
                section_iep_skills[skillset_id].remove(skill_id)
        


@implementer(IStudentIEP)
@adapter(IBasicPerson)
def getStudentIEP(person):
    annotations = IAnnotations(person)
    try:
        return annotations[STUDENT_IEP_KEY]
    except KeyError:
        iep = StudentIEP()
        iep.__parent__ = person
        annotations[STUDENT_IEP_KEY] = iep
        return iep



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Evolve database to generation 2.

Fix the mess made by evolution 1.
"""

from zope.annotation.interfaces import IAnnotations
from zope.app.generations.utility import findObjectsProviding, getRootFolder
from zope.component.hooks import getSite, setSite
from zope.keyreference.interfaces import IKeyReference

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.cando.course import getSectionSkills
from schooltool.cando.course import SectionSkill
from schooltool.cando.generations.evolve1 import pick_section

EVALUATIONS_KEY = "schooltool.evaluations"


def findSkill(score, section):
    section_worksheets = getSectionSkills(section)
    skill_name = score.requirement.__name__
    skillset_name = score.requirement.__parent__.__name__
    for section_skill_set in section_worksheets.values():
        for skill_key in section_skill_set.all_keys():
            skill = section_skill_set[skill_key]
            if (skill_name == skill.source_skill_name and
                skillset_name == skill.source_skillset_name):
                return skill
    return None


def findCandidateSection(score):
    courses = list(score.requirement.section.courses)

    evaluator_sections = []
    sections_attended = []
    for course in courses:
        for section in course.sections:
            instructors = [i.__name__ for i in section.instructors]
            if (score.evaluator in instructors and
                section not in evaluator_sections):
                evaluator_sections.append(section)
            if (score.evaluatee in section.members and
                section not in sections_attended):
                sections_attended.append(section)
    likely_sections = [section for section in sections_attended
                       if section in evaluator_sections]
    target_section = pick_section(score, likely_sections)
    if target_section is None:
        # Maybe evaulator has changed. Try picking sections
        # of this course attended by evaluatee
        target_section = pick_section(score, sections_attended)
    if target_section is None:
        # Try picking sections of this course taught by evaluator
        target_section = pick_section(score, evaluator_sections)
    if target_section is not None:
        return target_section
    return score.requirement.section


def isOlder(older_score, newer_score):
    if not older_score or not newer_score:
        return None
    return older_score.time <= newer_score.time


def replayScore(evaluations, score, skill):
    score.requirement = skill
    skill_ref = IKeyReference(skill)
    history = None

    if (evaluations._history is not None):
        history = evaluations._history.get(skill_ref)

    if (skill_ref not in evaluations._btree and
        not history):
        # No scores yet, so simply add it
        evaluations[skill] = score
        return

    # Our score is newer, so just set it
    latest_score = evaluations._btree.get(skill_ref)
    if isOlder(latest_score, score):
        evaluations[skill] = score
        return

    if not history:
        if not latest_score:
            evaluations[skill] = score
        else:
            # There is a newer score, but no other scores.  Just keep it.
            evaluations.appendToHistory(skill, score)
        return

    insert_before = None

    for index, historical_score in enumerate(history):
        if (historical_score is not None and
            not isOlder(historical_score, score)):
            # The score in history is newer than our misplaced score
            insert_before = index
            break

    # Score may want to be on top of history
    if insert_before is None:
        if not latest_score:
            evaluations[skill] = score
        else:
            evaluations.appendToHistory(skill, score)
        return

    evaluations._history[skill_ref].insert(insert_before, score)


def fixSkillScores(app, person):
    annotations = IAnnotations(person)
    if EVALUATIONS_KEY not in annotations:
        return
    evaluations = annotations[EVALUATIONS_KEY]

    misplaced_scores = []
    refs = set(evaluations._history or []).union(set(evaluations._btree or []))

    for ref in refs:
        skill = ref()
        if not isinstance(skill, SectionSkill):
            continue
        if (evaluations._history is not None and
            ref in evaluations._history):
            remove_idx = []
            for n, score in enumerate(evaluations._history[ref]):
                if (score is None or
                    not isinstance(score.requirement, SectionSkill)):
                    continue
                section = findCandidateSection(score)
                if section is score.requirement.section:
                    continue
                skill = findSkill(score, section)
                if skill is None:
                    continue
                misplaced_scores.append((score, skill))
                remove_idx.append(n)
            for n in reversed(remove_idx):
                evaluations._history[ref].pop(n)
        if ref in evaluations._btree:
            score = evaluations._btree[ref]
            if (score is not None and
                isinstance(score.requirement, SectionSkill)):
                section = findCandidateSection(score)
                if section is not score.requirement.section:
                    skill = findSkill(score, section)
                    if skill is not None:
                        misplaced_scores.append((score, skill))
                        del evaluations._btree[ref]
                        if (evaluations._history is not None and
                            ref in evaluations._history and
                            evaluations._history[ref]):
                            evaluations._btree[ref] = evaluations._history[ref].pop()

    for score, skill in misplaced_scores:
        replayScore(evaluations, score, skill)


def evolve(context):
    root = getRootFolder(context)

    old_site = getSite()
    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        setSite(app)

        for person in app['persons'].values():
            fixSkillScores(app, person)

    setSite(old_site)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2013 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Evolve database to generation 3.

Sets equivalent course skill for section skills.

Fixes a bug caused by the old skills importer (before rev 154) where
equivalent were removed for global skills, affecting also course
skills and breaking projects.
"""

from zope.annotation.interfaces import IAnnotations
from zope.app.generations.utility import findObjectsProviding, getRootFolder
from zope.component.hooks import getSite, setSite

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.relationships import CourseSections
from schooltool.relationship.relationship import relate

from schooltool.cando.course import COURSE_SKILLS_KEY
from schooltool.cando.course import SECTION_SKILLS_KEY
from schooltool.cando.skill import EquivalentSkills
from schooltool.cando.skill import URIEquivalent
from schooltool.cando.skill import URISkill


def getCourseSkillSets(course):
    course_annotations = IAnnotations(course)
    return course_annotations.get(COURSE_SKILLS_KEY, {})


def getSectionSkillSets(section):
    section_annotations = IAnnotations(section)
    return section_annotations.get(SECTION_SKILLS_KEY, {})


def evolveSectionSkills(section, course_skillsets):
    section_skillsets = getSectionSkillSets(section)
    for section_skillset in section_skillsets.values():
        course_skillset = course_skillsets[section_skillset.__name__]
        for section_skill in section_skillset.values():
            equivalent = list(EquivalentSkills.query(skill=section_skill))
            course_skill = course_skillset[section_skill.__name__]
            if course_skill not in equivalent:
                relate(URIEquivalent,
                       (section_skill, URISkill),
                       (course_skill, URISkill))


def evolve(context):
    root = getRootFolder(context)
    old_site = getSite()
    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        setSite(app)
        for cc in app['schooltool.course.course'].values():
            for course in cc.values():
                course_skillsets = getCourseSkillSets(course)
                if course_skillsets:
                    course_sections = list(CourseSections.query(course=course))
                    for section in course_sections:
                        evolveSectionSkills(section, course_skillsets)
    setSite(old_site)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Initial deployment of generations to schooltool.requirement.

Manually evolves to generation 1. Needed to support upgrade from a
database that did not have generations.
"""

import evolve1
import evolve2


def evolve(context):
    evolve1.evolve(context)
    # Do not run evolve 2 on install, because it fixes data
    # that were misplaced by older version of evolve1.
    # Now evolve1 does the job.
    #evolve2.evolve(context)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Document views.
"""

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, getUtility, getMultiAdapter
from zope.container.interfaces import INameChooser
from zope.interface import implements, directlyProvides
from zope.intid.interfaces import IIntIds
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL

import zc.table.column
import z3c.form.form
import z3c.form.button
import z3c.form.field

from schooltool.skin import flourish
from schooltool import table
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.app.browser.app import EditRelationships
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.app import ContentTitle
from schooltool.common.inlinept import InlineViewPageTemplate, InheritTemplate
from schooltool.schoolyear.interfaces import ISchoolYearContainer

from schooltool.cando.browser.model import LayersTable, LayerView, LayerEditView
from schooltool.cando.browser.model import EditChildLayersView
from schooltool.cando.browser.skill import SkillAddView, SkillView
from schooltool.cando.browser.skill import SkillSetEditView, SkillEditView
from schooltool.cando.interfaces import ILayerContainer, ILayer
from schooltool.cando.interfaces import INodeContainer, INode
from schooltool.cando.interfaces import IDocumentContainer, IDocument
from schooltool.cando.interfaces import ISkillSetContainer, ISkillSet
from schooltool.cando.model import LayerContainer, Layer, LayerLink
from schooltool.cando.model import NodeContainer, Node, NodeLink
from schooltool.cando.model import DocumentContainer, Document
from schooltool.cando.skill import SkillSetContainer, SkillSet, Skill

from schooltool.cando import CanDoMessage as _



class DocumentContainerAbsoluteURLAdapter(BrowserView):
    adapts(IDocumentContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = absoluteURL(app, self.request)
        return url + '/documents'

    __call__ = __str__


class DocumentsView(flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/view/container/table" />
      <h3 tal:condition="python: not len(context)" i18n:domain="schooltool">There are no documents.</h3>
    ''')

    @Lazy
    def container(self):
        return IDocumentContainer(ISchoolToolApplication(None))


class DocumentsTable(table.ajax.Table):

    def columns(self):
        default = table.ajax.Table.columns(self)
        return default

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})


class DocumentsTertiaryNavigationManager(
    flourish.page.TertiaryNavigationManager):

    template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    @property
    def items(self):
        tabs = (
            ((DocumentContainer, Document), 'documents', _('Documents')),
            ((LayerContainer, Layer), 'layers', _('Layers')),
            ((SkillSetContainer, SkillSet, Skill), 'skills', _('Skill Sets')),
            ((NodeContainer, Node), 'nodes', _('Search')),
            )
        result = []
        app = ISchoolToolApplication(None)
        for context_list, link, title in tabs:
            url = '%s/%s' % (absoluteURL(app, self.request), link)
            if 'Document' in str(self.view):
                active = (link == 'documents')
            else:
                active = (self.context.__class__ in context_list)
            result.append({
                'class': active and 'active' or None,
                'viewlet': u'<a href="%s">%s</a>' % (url, title),
                })
        return result


class DocumentsAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links in DocumentsView"""


class DocumentsActionsLinks(flourish.page.RefineLinksViewlet):
    """Manager for Actions links in DocumentsView"""


class ImportSkillsLinkViewlet(flourish.page.LinkViewlet):

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        app = ISchoolToolApplication(None)
        return "%s/%s" % (absoluteURL(app, self.request),
                          self.link)


class DocumentAddView(flourish.form.AddForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Document Information')
    fields = z3c.form.field.Fields(IDocument).select('title', 'description')

    @property
    def layer_titles(self):
        items = []
        for key in [k for k in self.request.keys() if k.startswith('row')]:
            index = int(key[3:])
            items.append([index, self.request[key]])
        return [v for k, v in reversed(sorted(items))]

    @property
    def nonempty_titles(self):
        return [t for t in self.layer_titles if t]

    def rows(self):
        titles = self.layer_titles
        if not titles:
            titles = [_('Skill Set'), _('Skill')]
        num_titles = len(titles)
        rows = []
        for index, title in enumerate(titles):
            rows.append({
                'name': 'row%d' % (num_titles - index - 1),
                'value': title,
                'add': index == 0,
                })
        return rows

    def updateActions(self):
        super(DocumentAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(DocumentAddView, self).handleAdd.func(self, action)

    @z3c.form.button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        app = ISchoolToolApplication(None)
        url = '%s/documents' % absoluteURL(app, self.request)
        self.request.response.redirect(url)

    def create(self, data):
        document = Document(data['title'])
        z3c.form.form.applyChanges(self, document, data)
        return document

    def add(self, document):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', document)
        self.context[name] = document
        self._document = document

        layers = ILayerContainer(ISchoolToolApplication(None))
        chooser = INameChooser(layers)
        previous_layer = None
        for title in self.nonempty_titles:
            layer = Layer(title)
            name = chooser.chooseName(u'', layer)
            layers[name] = layer
            if previous_layer is not None:
                layer.parents.add(previous_layer)
            document.hierarchy.add(layer)
            previous_layer = layer
        return document

    def nextURL(self):
        return absoluteURL(self._document, self.request)


class DocumentMixin(object):

    def is_document(self):
        return IDocument(self.context, None) is not None

    def get_document(self):
        return self.context

    def get_node(self):
        return None

    def get_layer_hierarchy(self):
        document = self.get_document()
        if document is None:
            return []
        return document.getOrderedHierarchy()

    def get_layer(self):
        layer_id = self.request.get('layer', '')
        app = ISchoolToolApplication(None)
        return ILayerContainer(app).get(layer_id, None)

    def is_skillset_layer(self):
        return len(self.get_layer_hierarchy()) < 3

    def get_next_layer(self):
        hierarchy = self.get_layer_hierarchy()
        if not hierarchy:
            return None
        return hierarchy[0]

    def get_previous_layer(self):
        return None

    def get_children(self):
        if INode(self.context, None) is not None:
            return sorted(NodeLink.query(parent=self.context),
                          key=lambda l: l.__name__)
        return []

    @property
    def layer_title(self):
        layer = self.get_layer()
        if layer is None:
            return _('Skill Set')
        return layer.title

    @property
    def next_layer_title(self):
        layer = self.get_next_layer()
        if layer is None:
            return _('Skill Set')
        return layer.title

    def build_query_string(self, **kw):
        query_string_dict = {}
        document = self.get_document()
        if document is not None:
            query_string_dict['document'] = document.__name__
        layer = kw.get('layer', None)
        if layer is not None:
            query_string_dict['layer'] = layer.__name__
        node = kw.get('node', None)
        if node is not None:
            query_string_dict['node'] = node.__name__

        query_string = ''
        for index, (k, v) in enumerate(query_string_dict.items()):
            query_string += index and '&' or '?'
            query_string += '%s=%s' % (k, v)
        return query_string

    def make_node_item(self, node):
        query_string = self.build_query_string(layer=self.get_next_layer())
        return {
            'url': '%s/document.html%s' % (absoluteURL(node, self.request),
                                           query_string),
            'obj': node,
            }

    def make_skillset_item(self, skillset):
        query_string = self.build_query_string(layer=self.get_next_layer(),
                                               node=self.get_node())
        return {
            'url': '%s/document.html%s' % (absoluteURL(skillset, self.request),
                                           query_string),
            'obj': skillset,
            }

    @property
    def items(self):
        result = []
        if self.is_skillset_layer():
            for skillset in self.context.skillsets:
                result.append(self.make_skillset_item(skillset))
        else:
            for node in self.get_children():
                result.append(self.make_node_item(node))
        return result

    @property
    def add_url(self):
        if self.is_skillset_layer():
            url = 'add_document_skillset.html'
            query_string = self.build_query_string(layer=self.get_next_layer(),
                                                   node=self.get_node())
        else:
            url = 'add_document_node.html'
            query_string = self.build_query_string(layer=self.get_next_layer())
        return '%s%s' % (url, query_string)


class DocumentNodeMixin(DocumentMixin):

    def get_document(self):
        document_id = self.request.get('document', '')
        app = ISchoolToolApplication(None)
        return IDocumentContainer(app).get(document_id, None)

    def get_node(self):
        if INode(self.context, None) is not None:
            return self.context
        node_id = self.request.get('node', '')
        app = ISchoolToolApplication(None)
        return INodeContainer(app).get(node_id, None)

    def is_skillset_layer(self):
        hierarchy = self.get_layer_hierarchy()
        if len(hierarchy) < 3:
            return True
        current_layer = self.get_next_layer()
        if current_layer is None:
            return True
        for index, layer in enumerate(hierarchy):
            if layer is current_layer and index < len(hierarchy) - 2:
                return False
        return True

    def get_next_layer(self):
        hierarchy = self.get_layer_hierarchy()
        if hierarchy:
            current_layer = self.get_layer()
            if current_layer is None:
                return None
            for index, layer in enumerate(hierarchy):
                if layer is current_layer and index < len(hierarchy) - 1:
                    return hierarchy[index + 1]
        return None

    def get_previous_layer(self):
        hierarchy = self.get_layer_hierarchy()
        if hierarchy:
            current_layer = self.get_layer()
            if current_layer is None:
                return None
            for index, layer in enumerate(hierarchy):
                if layer is current_layer and index > 0:
                    return hierarchy[index - 1]
        return None


class DocumentSkillSetMixin(DocumentNodeMixin):

    @property
    def next_layer_title(self):
        hierarchy = self.get_layer_hierarchy()
        if len(hierarchy) > 1:
            return hierarchy[-1].title
        return _('Skill')


class DocumentSkillMixin(DocumentSkillSetMixin):

    @property
    def layer_title(self):
        layer = self.get_layer()
        if layer is None:
            return _('Skill')
        return layer.title


class DocumentAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links in DocumentView"""


class DocumentEditLinks(flourish.page.RefineLinksViewlet):
    """Manager for Edit links in DocumentView"""


class DocumentAddNodeLink(flourish.page.LinkViewlet, DocumentMixin):

    @property
    def title(self):
        return self.next_layer_title

    @property
    def url(self):
        return self.add_url


class DocumentView(flourish.form.DisplayForm, DocumentMixin):

    template = InheritTemplate(flourish.page.Page.template)
    label = None

    fields = z3c.form.field.Fields(IDocument).select('description')

    @property
    def legend(self):
        return _('${layer} list',
                 mapping={'layer': self.next_layer_title})

    @property
    def can_edit(self):
        return flourish.canEdit(self.context)

    @property
    def edit_url(self):
        return absoluteURL(self.context, self.request) + '/edit.html'

    @property
    def done_link(self):
        app = ISchoolToolApplication(None)
        return '%s/documents' % absoluteURL(app, self.request)


def document_node_title_formatter(value, item, formatter):
    return '<a href="%s">%s</a>' % (item['url'], value)


class DocumentNodesTable(table.ajax.Table):

    batch_size = 0

    def items(self):
        return self.view.items

    def columns(self):
        label = table.column.LocaleAwareGetterColumn(
            name='label',
            title=_('Label'),
            getter=lambda i, f: i['obj'].label or '',
            subsort=True)
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_('Title'),
            getter=lambda i, f: i['obj'].title,
            cell_formatter=document_node_title_formatter,
            subsort=True)
        return [label, title]

    def sortOn(self):
        return (('label', False), ('title', False))

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})


class DocumentNodesTableParameters(flourish.viewlet.Viewlet):

    parameters = ['document', 'layer', 'node']

    template = InlineViewPageTemplate('''
      <tal:block repeat="parameter view/parameters">
        <input type="hidden" tal:define="value python:request.get(parameter)"
               tal:condition="value"
               tal:attributes="name parameter;
                               value value;"
               />
      </tal:block>
    ''')


class DocumentEditView(flourish.form.Form, z3c.form.form.EditForm):
    fields = z3c.form.field.Fields(IDocument)
    fields = fields.select('title', 'description')

    legend = _('Document')

    def applyChanges(self, data):
        if data['description'] is None:
            data['description'] = u''
        super(DocumentEditView, self).applyChanges(data)

    @z3c.form.button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(DocumentEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @z3c.form.button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(DocumentEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class EditDocumentHierarchyView(EditRelationships):
    current_title = _("Current document hierarchy layers")
    available_title = _("Available layers")

    def getCollection(self):
        return self.context.getOrderedHierarchy()

    def getAvailableItemsContainer(self):
        return ILayerContainer(ISchoolToolApplication(None))

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]


class LayerContainerSourceMixin(object):

    @property
    def source(self):
        return ILayerContainer(ISchoolToolApplication(None))

    def sortOn(self):
        return []


class AvailableLayersTable(LayerContainerSourceMixin,
                           RelationshipAddTableMixin,
                           LayersTable):
    pass


class RemoveLayersTable(LayerContainerSourceMixin,
                        RelationshipRemoveTableMixin,
                        LayersTable):
    pass


class DocumentLayerView(LayerView, DocumentNodeMixin):

    @property
    def edit_url(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string()
        return '%s/edit_document_layer.html%s' % (url, query_string)

    @property
    def edit_children_url(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string()
        return '%s/edit_document_layer_children.html%s' % (url, query_string)

    @property
    def done_link(self):
        document = self.get_document()
        if document is None:
            app = ISchoolToolApplication(None)
            return '%s/documents' % absoluteURL(app, self.request)
        return absoluteURL(document, self.request)


class DocumentLayerEditView(LayerEditView, DocumentNodeMixin):

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string()
        return '%s/document.html%s' % (url, query_string)


class EditDocumntLayerChildrenView(EditChildLayersView, DocumentNodeMixin):

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string()
        return '%s/document.html%s' % (url, query_string)


class DocumentNodeView(flourish.form.DisplayForm, DocumentNodeMixin):
    """Same as DocumentView but for a particular node"""

    template = InheritTemplate(flourish.page.Page.template)
    label = None

    fields = z3c.form.field.Fields(INode).select('title', 'description',
                                                 'label')

    @property
    def subtitle(self):
        return _('View ${layer}',
                 mapping={'layer': self.layer_title})

    @property
    def legend(self):
        return _('${layer} list',
                 mapping={'layer': self.next_layer_title})

    @property
    def can_edit(self):
        return flourish.canEdit(self.context)

    @property
    def edit_url(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_layer())
        return '%s/edit_document_node.html%s' % (url, query_string)

    @property
    def done_link(self):
        document = self.get_document()
        if document is None:
            app = ISchoolToolApplication(None)
            return '%s/documents' % absoluteURL(app, self.request)
        layer = self.get_previous_layer()
        if layer is not None:
            for parent in self.context.parents:
                if layer in parent.layers:
                    url = absoluteURL(parent, self.request)
                    query_string = self.build_query_string(layer=layer)
                    return '%s/document.html%s' % (url, query_string)
        return absoluteURL(document, self.request)


class DocumentNodeAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links in DocumentView"""


class DocumentNodeAddNodeLink(flourish.page.LinkViewlet, DocumentNodeMixin):

    @property
    def title(self):
        return self.next_layer_title

    @property
    def url(self):
        return self.add_url


class DocumentAddNodeBase(flourish.form.AddForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    fields = z3c.form.field.Fields(INode).select('title', 'description',
                                                 'label')

    @property
    def subtitle(self):
        return _('Add ${layer}',
                 mapping={'layer': self.layer_title})

    @property
    def legend(self):
        return _('${layer} Information',
                 mapping={'layer': self.layer_title})

    def updateActions(self):
        super(DocumentAddNodeBase, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(DocumentAddNodeBase, self).handleAdd.func(self, action)

    @z3c.form.button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def create(self, data):
        if data['description'] is None:
            data['description'] = u''
        node = Node(data['title'])
        z3c.form.form.applyChanges(self, node, data)
        node.parents.add(removeSecurityProxy(self.context))
        layer = self.get_layer()
        if layer is not None:
            node.layers.add(removeSecurityProxy(layer))
        return node

    def add(self, node):
        nodes = INodeContainer(ISchoolToolApplication(None))
        chooser = INameChooser(nodes)
        name = chooser.chooseName(u'', node)
        nodes[name] = node
        return node


class DocumentAddNodeView(DocumentAddNodeBase, DocumentMixin):
    """Add Node from DocumentView"""

    @property
    def title(self):
        return _('Skills Document')

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class DocumentNodeAddNodeView(DocumentAddNodeBase, DocumentNodeMixin):
    """Add Node from DocumentNodeView"""

    @property
    def title(self):
        return self.context.title

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/document.html'


class DocumentNodeEditView(flourish.form.Form, z3c.form.form.EditForm,
                           DocumentNodeMixin):
    fields = z3c.form.field.Fields(INode)
    fields = fields.select('title', 'description', 'label')

    legend = _('Change information')

    def applyChanges(self, data):
        if data['description'] is None:
            data['description'] = u''
        super(DocumentNodeEditView, self).applyChanges(data)

    @z3c.form.button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(DocumentNodeEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @z3c.form.button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def updateActions(self):
        super(DocumentNodeEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/document.html'

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_layer())
        return '%s/document.html%s' % (url, query_string)


class DocumentAddSkillSetBase(flourish.form.AddForm):

    _skillset = None
    label = None
    fields = z3c.form.field.Fields(ISkillSet)
    fields = fields.select('title', 'description', 'label')

    @property
    def legend(self):
        return self.layer_title

    @property
    def subtitle(self):
        return _('Add ${layer}',
                 mapping={'layer': self.layer_title})

    def updateActions(self):
        super(DocumentAddSkillSetBase, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def create(self, data):
        skillset = SkillSet(data['title'])
        z3c.form.form.applyChanges(self, skillset, data)
        self._skillset = skillset
        return skillset

    def add(self, skillset):
        skillsets = ISkillSetContainer(ISchoolToolApplication(None))
        chooser = INameChooser(skillsets)
        name = unicode(skillset.title).encode('punycode')
        name = name[:8]+str(len(skillsets)+1)
        name = chooser.chooseName(name, skillset)
        skillsets[name] = skillset
        removeSecurityProxy(self.context.skillsets).add(
            removeSecurityProxy(skillset))
        return skillset

    def nextURL(self):
        if self._skillset is not None:
            url = absoluteURL(self._skillset, self.request) + '/document.html'
            layer = self.get_layer()
        else:
            url = self.contextURL()
            layer = self.get_previous_layer()
        query_string = self.build_query_string(layer=layer,
                                               node=self.get_node())
        return '%s%s' % (url, query_string)


class DocumentAddSkillSetView(DocumentAddSkillSetBase, DocumentMixin):
    """Add SkillSet from DocumentView"""

    def contextURL(self):
        return absoluteURL(self.context, self.request)


class DocumentNodeAddSkillSetView(DocumentAddSkillSetBase, DocumentNodeMixin):
    """Add SkillSet from DocumentNodeView"""

    def contextURL(self):
        return absoluteURL(self.context, self.request) + '/document.html'


class DocumentSkillSetView(flourish.form.DisplayForm, DocumentSkillSetMixin):
    template = InheritTemplate(flourish.page.Page.template)
    fields = z3c.form.field.Fields(ISkillSet)
    fields = fields.select('description', 'label', 'retired')

    @property
    def subtitle(self):
        return _('View ${layer}',
                 mapping={'layer': self.layer_title})

    @property
    def legend(self):
        layer = self.get_next_layer()
        if layer is None:
            return _('Skills')
        return _('${layer} list',
                 mapping={'layer': layer.title})

    @property
    def can_edit(self):
        return flourish.canEdit(self.context)

    @property
    def edit_url(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_layer(),
                                               node=self.get_node())
        return '%s/edit_document_skillset.html%s' % (url, query_string)

    @property
    def done_link(self):
        app = ISchoolToolApplication(None)
        document = self.get_document()
        if document is None:
            return '%s/skills' % absoluteURL(app, self.request)
        node_id = self.request.get('node', '')
        node = INodeContainer(app).get(node_id, None)
        if node is not None:
            previous_layer = self.get_previous_layer()
            query_string = self.build_query_string(layer=previous_layer)
            return '%s/document.html%s' % (absoluteURL(node, self.request),
                                           query_string)
        else:
            return '%s/index.html' % absoluteURL(document, self.request)


class DocumentSkillSetSkillTable(table.ajax.Table, DocumentSkillSetMixin):

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})

    def sortOn(self):
        return (("label", False), ("title", False))

    def title_url_formatter(self, value, item, formatter):
        query_string = self.build_query_string(layer=self.get_next_layer(),
                                               node=self.get_node())
        url = '%s/document.html%s' % (absoluteURL(item, formatter.request),
                                      query_string)
        return '<a href="%s">%s</a>' % (url, value)

    def columns(self):
        title = zc.table.column.GetterColumn(
            name='title',
            title=_(u"Title"),
            cell_formatter=lambda v, i, f: self.title_url_formatter(v, i, f),
            getter=lambda i, f: i.title,
            subsort=True)
        directlyProvides(title, zc.table.interfaces.ISortableColumn)
        required = zc.table.column.GetterColumn(
            name='required',
            title=_(u'Required'),
            getter=lambda i, f: i.required and _('Yes') or _('No'))
        directlyProvides(required, zc.table.interfaces.ISortableColumn)
        label = table.column.LocaleAwareGetterColumn(
            name='label',
            title=_(u'Label'),
            getter=lambda i, f: i.label or '',
            subsort=True)
        return [label, title, required]


class DocumentSkillSetEditView(SkillSetEditView, DocumentSkillSetMixin):

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_layer(),
                                               node=self.get_node())
        return '%s/document.html%s' % (url, query_string)


class DocumentSkillSetLinks(flourish.page.RefineLinksViewlet):
    pass


class DocumentAddSkillLink(flourish.page.LinkViewlet, DocumentSkillSetMixin):

    @property
    def title(self):
        return self.next_layer_title

    @property
    def url(self):
        url = 'add_document_skill.html'
        query_string = self.build_query_string(layer=self.get_next_layer(),
                                               node=self.get_node())
        return '%s%s' % (url, query_string)


class DocumentAddSkillView(SkillAddView, DocumentSkillSetMixin):

    @property
    def legend(self):
        return _('${layer} Information',
                 mapping={'layer': self.next_layer_title})

    @property
    def subtitle(self):
        return _('Add ${layer}',
                 mapping={'layer': self.next_layer_title})

    def create(self, data):
        skill = Skill(data['title'])
        z3c.form.form.applyChanges(self, skill, data)
        self._skill = skill
        return skill

    def add(self, skill):
        skillset = self.context
        if not skill.label:
            skill.label = u'%02d' % (len(skillset) + 1)
        chooser = INameChooser(skillset)
        if skill.external_id:
            name = skill.external_id
        else:
            name = unicode(skill.title).encode('punycode')
            name = name[:8]+str(len(skillset)+1)
        name = chooser.chooseName(name, skill)
        skillset[name] = skill
        return skill

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        if self.add_next:
            query_string = self.build_query_string(layer=self.get_layer(),
                                                   node=self.get_node())
            return '%s/add_document_skill.html%s' % (url, query_string)
        else:
            previous_layer = self.get_previous_layer()
            query_string = self.build_query_string(layer=previous_layer,
                                                   node=self.get_node())
            return '%s/document.html%s' % (url, query_string)


class DocumentSkillView(SkillView, DocumentSkillMixin):

    @property
    def subtitle(self):
        return _('View ${layer}',
                 mapping={'layer': self.layer_title})

    @property
    def can_edit(self):
        return flourish.canEdit(self.context)

    @property
    def edit_url(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_layer(),
                                               node=self.get_node())
        return '%s/edit_document_skill.html%s' % (url, query_string)

    @property
    def done_link(self):
        url = absoluteURL(self.context.__parent__, self.request)
        query_string = self.build_query_string(layer=self.get_previous_layer(),
                                               node=self.get_node())
        return '%s/document.html%s' % (url, query_string)


class DocumentSkillEditView(SkillEditView, DocumentSkillMixin):

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_layer(),
                                               node=self.get_node())
        return '%s/document.html%s' % (url, query_string)


class CustomizeDocumentLink(flourish.page.LinkViewlet):
    pass




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Skills document model."""
import itertools

from persistent import Persistent
import zope.lifecycleevent
from zope.container.btree import BTreeContainer
from zope.container.contained import Contained
from zope.component import adapts, adapter
from zope.interface import implements, implementer
from zope.security.proxy import removeSecurityProxy

from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.catalog import AttributeCatalog
from zope.index.text.interfaces import ISearchableText
from schooltool.cando import interfaces
from schooltool.cando.skill import SearchableTextMixin
from schooltool.cando.skill import URISkillSet
from schooltool.cando.skill import setSearchableIndexes
from schooltool.relationship import URIObject
from schooltool.relationship import RelationshipSchema, RelationshipProperty
from schooltool.relationship.interfaces import InvalidRelationship
from schooltool.relationship.interfaces import IBeforeRelationshipEvent
from schooltool.relationship.interfaces import IBeforeRemovingRelationshipEvent
from schooltool.table.catalog import ConvertingSetIndex


URILayer = URIObject(
    'http://schooltool.org/ns/cando/model/layer',
    'Layer',
    'A layer of the model.')

URIParentLayer = URIObject(
    'http://schooltool.org/ns/cando/model/parent_layer',
    'Parent layer',
    'A parent layer.')

URILayerLink = URIObject(
    'http://schooltool.org/ns/cando/model/layer_link',
    'Layer link',
    'A link between two layers in the model.')

LayerLink = RelationshipSchema(URILayerLink,
                              parent=URIParentLayer,
                              child=URILayer)


URINodeLink = URIObject(
    'http://schooltool.org/ns/cando/model/node_link',
    'Node link',
    'A link between two nodes in the model.')

URINode = URIObject(
    'http://schooltool.org/ns/cando/model/node',
    'Node',
    'A node.')

URIParentNode = URIObject(
    'http://schooltool.org/ns/cando/model/parent_node',
    'Parent node',
    'A parent node.')

NodeLink = RelationshipSchema(URINodeLink,
                              parent=URIParentNode,
                              child=URINode)

URINodeLayer = URIObject(
    'http://schooltool.org/ns/cando/model/node_layer',
    'Node layer',
    'A model layer the node is in.')

NodeLayer = RelationshipSchema(URINodeLayer,
                               node=URINode,
                               layer=URILayer)


URINodeSkillSets = URIObject(
    'http://schooltool.org/ns/cando/model/node_skillsets',
    'Node skillsets',
    'Skillsets this model node implements.')

NodeSkillSets = RelationshipSchema(
    URINodeSkillSets,
    node=URINode,
    skillset=URISkillSet)
# XXX: order in extra_info if neccessary


URIDocument = URIObject(
    'http://schooltool.org/ns/cando/model/document',
    'Document',
    'A document.')

URIDocumentHierarchy = URIObject(
    'http://schooltool.org/ns/cando/model/document_hierarchy',
    'Document hierarchy',
    'A model layer that is in the document hierarchy.')

DocumentHierarchy = RelationshipSchema(URIDocumentHierarchy,
                                       document=URIDocument,
                                       layer=URILayer)


class LayerContainer(BTreeContainer):
    """Container of layers."""
    implements(interfaces.ILayerContainer)


class Layer(Persistent, Contained):
    implements(interfaces.ILayerContained)

    title = None
    parents = RelationshipProperty(URILayerLink, URILayer, URIParentLayer)
    children = RelationshipProperty(URILayerLink, URIParentLayer, URILayer)

    def __init__(self, title):
        self.title = title

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.title)


@adapter(ISchoolToolApplication)
@implementer(interfaces.ILayerContainer)
def getLayerContainer(app):
    return app['schooltool.cando.layer']


class LayerAppInit(InitBase):

    def __call__(self):
        self.app['schooltool.cando.layer'] = LayerContainer()


class LayerStartUp(StartUpBase):

    def __call__(self):
        if 'schooltool.cando.layer' not in self.app:
            self.app['schooltool.cando.layer'] = LayerContainer()


class NodeContainer(BTreeContainer):
    """Container of nodes."""
    implements(interfaces.INodeContainer)


class Node(Persistent, Contained):
    implements(interfaces.INodeContained)

    title = u''
    description = u''
    label = u''
    retired = False
    layers = RelationshipProperty(URINodeLayer, URINode, URILayer)
    parents = RelationshipProperty(URINodeLink, URINode, URIParentNode)
    children = RelationshipProperty(URINodeLink, URIParentNode, URINode)
    skillsets = RelationshipProperty(URINodeSkillSets, URINode, URISkillSet)

    def __init__(self, title=u'', description=u'', label=u'', retired=False):
        self.title = title
        self.description = description
        self.label = label
        self.retired = retired

    def findPaths(self):
        paths = [(self,)]
        expanded = True
        while expanded:
            expanded = False
            prev_paths = paths
            paths = []
            for path in prev_paths:
                parents = sorted(list(path[0].parents), key=lambda n:n.__name__)
                if parents:
                    for parent in parents:
                        paths.append((parent,)+path)
                        expanded = True
                else:
                    paths.append(path)
        return paths

    def __repr__(self):
        return '<%s %r %s>' % (self.__class__.__name__, self.title,
                               ', '.join([str(l) for l in self.layers]))


@adapter(ISchoolToolApplication)
@implementer(interfaces.INodeContainer)
def getNodeContainer(app):
    return app['schooltool.cando.node']


class NodeAppInit(InitBase):

    def __call__(self):
        self.app['schooltool.cando.node'] = NodeContainer()


class NodeStartUp(StartUpBase):

    def __call__(self):
        if 'schooltool.cando.node' not in self.app:
            self.app['schooltool.cando.node'] = NodeContainer()


class CyclicRelationship(InvalidRelationship):

    def __init__(self, child, parent, parents):
        self.child = child
        self.parent = parent
        self.parents = parents
        InvalidRelationship.__init__(self, child, parent, parents)

    def __str__(self):
        return '%s' % (
            '\n'.join([
                    'child %s' % self.child,
                    'target parent %s' % self.parent,
                    'distant parents %s' % ', '.join(
                        sorted([str(l) for l in self.parents]))
                    ]))


def _expand_nodes(nodes, functor, recursive=True):
    nodes = set(nodes)
    open = set(nodes)
    parents = set()
    while open:
        p = open.pop()
        if p not in parents:
            if recursive or p in nodes:
                open.update(set(functor(p)).difference(parents))
            if p not in nodes:
                parents.add(p)
    return parents


@adapter(IBeforeRelationshipEvent)
def preventLayerCycles(event):
    match = event.match(LayerLink)
    if match is None:
        match = event.match(NodeLink)
    if match is None:
        return
    parent = removeSecurityProxy(match.parent)
    child = removeSecurityProxy(match.child)
    # Explicitly don't handle existing child-parent relationships,
    # only parent loops.  So, check only parents of parent node.
    parents = _expand_nodes(nodes=[parent], functor=lambda n: n.parents)
    if child in parents:
        raise CyclicRelationship(child, parent, parents)


class ViolateLayerModel(InvalidRelationship):
    pass


def validateAgainstNode(candidate, distant_parent, valid_layers):
    open = set([distant_parent])
    while open:
        # Walk up to parents with layers and check if our layers are valid
        with_no_layers = set()
        for model_node in open:
            node_layers = list(model_node.layers)
            if node_layers:
                for layer in node_layers:
                    if layer not in valid_layers:
                        raise ViolateLayerModel('\n'.join([
                            'candidate %s' % candidate,
                            'acceptable parent layers %s' % ', '.join(
                                sorted([str(l) for l in valid_layers])),
                            'violates parent %s' % model_node]))
            else:
                with_no_layers.add(model_node)
        open = _expand_nodes(nodes=with_no_layers,
                             functor=lambda n:n.parents,
                             recursive=False)


@adapter(IBeforeRelationshipEvent)
def nodeLinkDoesntViolateModel(event):
    match = event.match(NodeLink)
    if match is None:
        return
    parent = removeSecurityProxy(match.parent)
    child = removeSecurityProxy(match.child)

    child_layers = set(child.layers)
    if child_layers:
        valid_layers = _expand_nodes(child_layers,
                                     functor=lambda n: n.parents,
                                     recursive=False)
        validateAgainstNode(child, parent, valid_layers)

    def expand_children_if_without_layers(node):
        if node.layers:
            return []
        return NodeLink.query(parent=node)

    check_children = _expand_nodes(nodes=[child],
                                   functor=expand_children_if_without_layers)
    for node in check_children:
        node_layers = set(node.layers)
        if node_layers:
            valid_layers = set(itertools.chain(*[n.parents for n in node_layers]))
            validateAgainstNode(node, parent, valid_layers)


class InvalidLayerLink(InvalidRelationship):

    def __init__(self, layer=None, node=None, offends_node=None):
        self.layer = layer
        self.node = node
        self.offends_node = offends_node
        InvalidRelationship.__init__(
            self, layer, node, offends_node)

    def __str__(self):
        return '%s' % (
            '\n'.join([
                    'setting layer %s' % self.layer,
                    'for node %s' % self.node,
                    'offends node %s' % self.offends_node,
                    ]))


def makeLayerValidator(functor, valid_set, invalid_set):
    def validator(node):
        node_layers = set(node.layers)
        offending_layers = node_layers.intersection(invalid_set)
        if offending_layers:
            raise InvalidLayerLink(offends_node=node)
        if not node_layers.isdisjoint(valid_set):
            # valid layer found, stop expanding
            return []
        return functor(node)
    return validator


@adapter(IBeforeRelationshipEvent)
def nodeLayerDoesntViolateModel(event):
    match = event.match(NodeLayer)
    if match is None:
        return
    node = removeSecurityProxy(match.node)
    layer = removeSecurityProxy(match.layer)

    get_layer_parents = lambda l:LayerLink.query(child=l)
    get_layer_children = lambda l:LayerLink.query(parent=l)

    immediate_parents = set(_expand_nodes(
            [layer], functor=get_layer_parents, recursive=False))
    distant_parents = set(_expand_nodes(
            immediate_parents, functor=get_layer_parents))
    all_parents = immediate_parents.union(distant_parents)

    immediate_children = set(_expand_nodes(
            [layer], functor=get_layer_children, recursive=False))
    distant_children = set(_expand_nodes(
            immediate_children, functor=get_layer_children))
    all_children = immediate_children.union(distant_children)

    valid_parents = immediate_parents
    invalid_parents = distant_parents.union(all_children)

    get_node_parents = lambda l:NodeLink.query(child=l)
    get_node_children = lambda l:NodeLink.query(parent=l)

    try:
        _expand_nodes([node], functor=makeLayerValidator(
                functor=get_node_parents,
                valid_set=valid_parents,
                invalid_set=invalid_parents,
                ))
    except InvalidLayerLink, e:
        raise InvalidLayerLink(
            layer=layer, node=node, offends_node=e.offends_node)

    valid_children = immediate_children
    invalid_children = distant_children.union(all_parents)

    try:
        _expand_nodes([node], functor=makeLayerValidator(
                functor=get_node_children,
                valid_set=valid_children,
                invalid_set=invalid_children,
                ))
    except InvalidLayerLink, e:
        raise InvalidLayerLink(
            layer=layer, node=node, offends_node=e.offends_node)


class CannotRemoveLayer(InvalidRelationship):

    def __init__(self, layer=None, node=None,
                 parent_nodes=None, child_nodes=None):
        self.layer = layer
        self.node = node
        self.parent_nodes = parent_nodes
        self.child_nodes = child_nodes
        InvalidRelationship.__init__(
            self, layer, node, parent_nodes, child_nodes)

    def __str__(self):
        return '%s' % (
            '\n'.join([
                    'setting layer %s' % self.layer,
                    'for node %s' % self.node,
                    'breaks parents %s' % ', '.join(
                        sorted([str(l) for l in self.parent_nodes])),
                    'and children %s' % ', '.join(
                        sorted([str(l) for l in self.child_nodes])),
                    ]))


def makeNodeWithLayersFinder(layers, functor, result):
    def find_node_with_layers(node):
        matched_layers = layers.intersection(set(node.layers))
        if matched_layers:
            result.append(node)
            return []
        return functor(node)
    return find_node_with_layers


@adapter(IBeforeRemovingRelationshipEvent)
def removingLayerDoesntViolateModel(event):
    match = event.match(NodeLayer)
    if match is None:
        return
    node = removeSecurityProxy(match.node)
    layer = removeSecurityProxy(match.layer)
    get_layer_parents = lambda l:LayerLink.query(child=l)
    get_layer_children = lambda l:LayerLink.query(parent=l)
    immediate_parents = set(_expand_nodes(
            [layer], functor=get_layer_parents, recursive=False))
    immediate_children = set(_expand_nodes(
            [layer], functor=get_layer_children, recursive=False))

    get_node_parents = lambda l:NodeLink.query(child=l)
    parent_nodes_in_model = []
    _expand_nodes(
        [node],
        functor=makeNodeWithLayersFinder(
            immediate_parents, get_node_parents, parent_nodes_in_model
            ))

    get_node_children = lambda l:NodeLink.query(parent=l)
    child_nodes_in_model = []
    _expand_nodes(
        [node],
        functor=makeNodeWithLayersFinder(
            immediate_children, get_node_children, child_nodes_in_model
            ))

    if parent_nodes_in_model and child_nodes_in_model:
        raise CannotRemoveLayer(
            layer=layer, node=node,
            parent_nodes=parent_nodes_in_model,
            child_nodes=child_nodes_in_model)


def nodeModifiedOnNodeLayerChange(event):
    match = event.match(NodeLayer)
    if match is None:
        return
    zope.lifecycleevent.modified(match.node)


def getOrderedByHierarchy(layers):
    layers = list(layers)
    orphans = []
    result = []
    for layer in layers:
        if not list(layer.parents) and not list(layer.children):
            orphans.append(layer)
        else:
            for index, item in enumerate(result):
                parents = _expand_nodes(nodes=[item], functor=lambda n: n.parents)
                if layer in parents:
                    result.insert(index, layer)
                    break
            else:
                result.append(layer)
    result.extend(sorted(orphans, key=lambda o: o.title))
    return result


class DocumentContainer(BTreeContainer):
    """Container of documents."""
    implements(interfaces.IDocumentContainer)


class Document(Node):
    implements(interfaces.IDocumentContained)

    hierarchy = RelationshipProperty(URIDocumentHierarchy, URIDocument,
                                     URILayer)

    def getOrderedHierarchy(self):
        return getOrderedByHierarchy(self.hierarchy)

    def __repr__(self):
        return '<%s %r %s>' % (self.__class__.__name__, self.title,
            ', '.join([str(l) for l in self.getOrderedHierarchy()]))


@adapter(ISchoolToolApplication)
@implementer(interfaces.IDocumentContainer)
def getDocumentContainer(app):
    return app['schooltool.cando.document']


class DocumentAppInit(InitBase):

    def __call__(self):
        self.app['schooltool.cando.document'] = DocumentContainer()


class DocumentStartUp(StartUpBase):

    def __call__(self):
        if 'schooltool.cando.document' not in self.app:
            self.app['schooltool.cando.document'] = DocumentContainer()


def get_node_layer_names(node):
    return [l.__name__ for l in node.layers]


def get_node_layer_titles(node):
    return [l.title for l in node.layers]


class NodeCatalog(AttributeCatalog):

    version = '1.3 - add __name__ to text index'
    interface = interfaces.INode
    attributes = ('title', 'label', 'description',
                  'required', 'retired')

    def setIndexes(self, catalog):
        super(NodeCatalog, self).setIndexes(catalog)
        setSearchableIndexes(catalog)
        catalog['layers'] = ConvertingSetIndex(converter=get_node_layer_names)
        catalog['layer_titles'] = ConvertingSetIndex(converter=get_node_layer_titles)


getNodeCatalog = NodeCatalog.get


class SearchableTextNode(SearchableTextMixin):

    adapts(interfaces.INode)
    implements(ISearchableText)

    def __init__(self, context):
        self.context = context

    def getSearchableText(self):
        result = [
            self.context.__name__,
            self.context.title,
            self.context.label or '',
            self.context.description or '',
            ]
        result.extend([l.title for l in self.context.layers])
        return ' '.join(result)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Evolve database to generation 1.

Sections get copies of skils instead of proxies.
"""
import transaction, gc

from zope.annotation.interfaces import IAnnotations
from zope.app.generations.utility import findObjectsProviding, getRootFolder
from zope.component import getUtility
from zope.component.hooks import getSite, setSite
from zope.intid.interfaces import IIntIds

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.relationships import CourseSections
from schooltool.cando.course import getSectionSkills
from schooltool.cando.course import SectionSkillSet, SectionSkill
from schooltool.term.interfaces import ITerm

COURSE_SKILLS_KEY = 'schooltool.cando.project.courseskills'
EVALUATIONS_KEY = "schooltool.evaluations"


def deploySkillSet(skillset, sections):
    int_ids = getUtility(IIntIds)
    for section in sections:
        worksheets = getSectionSkills(section)
        section_intid = int_ids.getId(section)
        worksheet = worksheets[skillset.__name__] = SectionSkillSet(skillset)
        for skill_name in skillset.all_keys():
            skill = skillset[skill_name]
            if skill_name not in worksheet.all_keys():
                target_skill = worksheet[skill_name] = SectionSkill(skill.title)
                target_skill.equivalent.add(skill)
            for attr in ('external_id', 'label', 'description',
                         'required', 'retired'):
                val = getattr(skill, attr, None)
                setattr(target_skill, attr, val)
            target_skill.section_intid = section_intid
            target_skill.source_skill_name = skill.__name__
            target_skill.source_skillset_name = skill.__parent__.__name__


def reassignScoreSkill(section_worksheets, evaluations, score):
    skill_name = score.requirement.__name__
    skillset_name = score.requirement.__parent__.__name__
    for section_skill_set in section_worksheets.values():
        for skill_key in section_skill_set.all_keys():
            skill = section_skill_set[skill_key]
            if (skill_name == skill.source_skill_name and
                skillset_name == skill.source_skillset_name):
                del evaluations[score.requirement]
                score.requirement = skill
                evaluations.addEvaluation(score)
                return


def pick_section(score, sections):
    date = score.time.date()
    before, after = [], []
    for section in sections:
        term = ITerm(section)
        if date in term:
            return section
        if date > term.last:
            before.append((term.last, section))
        if date < term.first:
            after.append((term.first, section))
    if before:
        return max(before)[1]
    if after:
        return min(after)[1]
    return None


def evolveCourse(connection, app, course):
    int_ids = getUtility(IIntIds)

    annotations = IAnnotations(course)
    if COURSE_SKILLS_KEY not in annotations:
        return
    courseskills = annotations[COURSE_SKILLS_KEY]
    if not courseskills:
        return
    course_sections = list(CourseSections.query(course=course))
    if not course_sections:
        return

    instructor_sections = {}
    student_sections = {}
    students = {}
    evaluations = {}
    for section in course_sections:
        sid = int_ids.getId(section)
        for student in section.members:
            if student not in students:
                students[student.__name__] = students
            annotations = IAnnotations(student)
            if EVALUATIONS_KEY in annotations:
                evaluations[student.__name__] = annotations[EVALUATIONS_KEY]
            if student.__name__ not in student_sections:
                student_sections[student.__name__] = {}
            student_sections[student.__name__][sid] = section
        for instructor in section.instructors:
            if instructor.__name__ not in instructor_sections:
                instructor_sections[instructor.__name__] = {}
            instructor_sections[instructor.__name__][sid] = section

    transaction.savepoint(True)

    for course_skillset in courseskills.values():
        deploySkillSet(course_skillset, course_sections)
        transaction.savepoint(True)
        connection.cacheMinimize()
        gc.collect()

    worksheet_cache = {}

    for course_skillset in courseskills.values():
        for skill_name in course_skillset.all_keys():
            skill = course_skillset[skill_name]
            for student_id, student_evaluations in evaluations.items():
                scores = student_evaluations.getEvaluationsForRequirement(skill)
                if not scores:
                    continue
                for score in scores.values():
                    sections_taught = instructor_sections.get(
                        score.evaluator, {})
                    sections_attended = student_sections.get(
                        score.evaluatee.__name__, {})

                    # Try picking sections of this course attended by evaluatee
                    # and taught by evaluator.
                    likely_sections = [
                        sections_attended[sid]
                        for sid in  set(sections_taught).intersection(sections_attended)]
                    target_section = pick_section(score, likely_sections)

                    if target_section is None:
                        # Maybe evaulator has changed. Try picking sections
                        # of this course attended by evaluatee
                        target_section = pick_section(score, sections_attended.values())
                    if target_section is None:
                        # Try picking sections of this course taught by evaluator
                        target_section = pick_section(score, sections_taught.values())
                    if target_section is None:
                        continue
                    if id(target_section) not in worksheet_cache:
                        worksheet_cache[id(target_section)] = getSectionSkills(target_section)
                    reassignScoreSkill(worksheet_cache[id(target_section)],
                                       student_evaluations, score)
            transaction.savepoint(True)
            connection.cacheMinimize()
            gc.collect()


def evolve(context):
    root = getRootFolder(context)

    old_site = getSite()
    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        setSite(app)

        for cc in app['schooltool.course.course'].values():
            for course in cc.values():
                evolveCourse(context.connection, app, course)

    setSite(old_site)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import zope.schema
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.container.interfaces import IContainer, IContained
from zope.container.constraints import contains
from zope.html.field import HtmlFragment
from zope.interface import Interface, Attribute, implements
from zope.schema import Choice

from schooltool.requirement.interfaces import IRequirement
from schooltool.gradebook.interfaces import IWorksheets, IWorksheet
from schooltool.gradebook.interfaces import IGradebook
from schooltool.gradebook.interfaces import IMyGrades
from schooltool.gradebook.interfaces import IStudentGradebook
from schooltool.cando import CanDoMessage as _


class ILabelTextLine(Interface):

    pass


class ISkillRequiredBool(Interface):

    pass


class LabelTextLine(zope.schema.TextLine):

    implements(ILabelTextLine)


class SkillRequiredBool(zope.schema.Bool):

    implements(ISkillRequiredBool)


class ILabel(Interface):

    label = LabelTextLine(
        title=_("Label"),
        description=_("Limit to 7 characters or less."),
        required=False,
        max_length=7,
        default=u'')


class ISkill(IRequirement, IAttributeAnnotatable, ILabel):

    external_id = zope.schema.TextLine(title=_("External ID"),
                                       required=False)
    required = SkillRequiredBool(title=_("Required?"))
    retired = zope.schema.Bool(title=_("Deprecated"),
                               description=_("Skill is no longer used"),
                               default=False)

    description = HtmlFragment(title=_("Description"), required=False)

    equivalent = Attribute("Directly equivalent skills.")

    scoresystem = Choice(
        title=_("Scoresystem"),
        description=_("The skill scoresystem."),
        vocabulary='schooltool.requirement.discretescoresystems',
        required=True)

    def findAllEquivalent():
        """Find all (including indirectly) equivalent skills."""

    def copy():
        """Return a copy of this skill."""


class ISkillSetContainer(IContainer):
    pass


class ISkillSet(IRequirement, IAttributeAnnotatable, ILabel):

    description = HtmlFragment(title=_("Description"), required=False)

    retired = zope.schema.Bool(title=_("Deprecated"),
                               description=_("Skill set is no longer used"),
                               default=False)


class ILayerContainer(IContainer):
    pass


class ILayer(Interface):
    title = zope.schema.TextLine(
        title=_("Title"))

    parents = Attribute("Parent layers")
    children = Attribute("Child layers")


class ILayerContained(ILayer, IContained, IAttributeAnnotatable):
    pass


class INodeContainer(IContainer):
    pass


class INode(ILabel):

    title = zope.schema.TextLine(
        title=_("Title"),
        required=True)
    description = HtmlFragment(
        title=_("Description"),
        required=False,
        default=u'')
    retired = zope.schema.Bool(
        title=_("Deprecated"),
        description=_("Node is no longer used"),
        default=False)

    layers = Attribute("Layers within this layer")
    parents = Attribute("Parent nodes")
    children = Attribute("Child nodes")
    skillsets = Attribute("Skill sets related to this node")

    def findPaths():
        """
          Return a list of paths (tuples) that lead
          (parent-to-child) to this node.
        """


class INodeContained(INode, IContained, IAttributeAnnotatable):
    pass


class IDocumentContainer(IContainer):
    pass


class IDocument(INode):

    hierarchy = Attribute("Hierarchy of layers for building node tree")

    def getOrderedHierarchy():
        """
          Return the ordered list of layers that represents the document
          hierarchy.
        """


class IDocumentContained(IDocument, IContained, IAttributeAnnotatable):
    pass


class IProject(ISkillSet, IWorksheet):
    pass


class IProjects(IWorksheets):
    contains('.IProject')


class ICourseProject(ISkillSet):
    """A template project."""
    contains('.ISkill')

    deployed = zope.schema.Bool(
        title=_("Project deployed"),
        description=_("Is this project deployed to course sections"),
        readonly=True,
        required=False
        )


class ICourseProjects(IRequirement):
    contains('.ICourseProject')

    def isDeployed(project, section):
        """Is given project deployed to that section?"""

    def deploy(self, key, section):
        """Deploy this project to that section."""


class ICourseSkills(IRequirement):
    contains('.ICourseSkillset')


class ICourseSkillSet(IContained):

    skillset = Attribute(u"The global skill set.")

    required = zope.schema.Dict(
        key_type=zope.schema.TextLine(title=u"Skill __name__ in skilset."),
        value_type=zope.schema.Bool(title=u"Is skill required"))

    retired = zope.schema.Dict(
        key_type=zope.schema.TextLine(title=u"Skill __name__ in skilset."),
        value_type=zope.schema.Bool(title=u"Retired skills should not be used."))


class ICourseSkill(ISkill):
    """Proxy for the real global skill"""

    course_skillset = Attribute(u"The course skill set.")


class ISectionSkills(IWorksheets):
    pass


class ISectionSkillSet(ISkillSet, IWorksheet):

    skillset = Attribute(u"The global skill set.")


class ISectionSkill(ISkill):
    section_intid = zope.schema.Int(title=_("Section intID"), required=False)
    section_skillset_name = zope.schema.Int(title=_("Skill set ID"), required=False)
    section_skill_name = zope.schema.Int(title=_("Skill ID"), required=False)


class ICanDoGradebook(IGradebook):
    pass


class IProjectsGradebook(ICanDoGradebook):
    pass


class ISkillsGradebook(ICanDoGradebook):
    pass


class IMySkillsGrades(IMyGrades):
    pass


class IMyProjectsGrades(IMyGrades):
    pass


class ICanDoStudentGradebook(IStudentGradebook):
    pass


class IStudentIEP(Interface):

    active = zope.schema.Bool(
        title=_('Active'),
        default=False)

    description = HtmlFragment(title=_("Description"), required=False)

    iep_skills = Attribute('IEP Skills')

    def getIEPSkills(section):
        """Returns IEP skills"""

    def addSkill(section, skill):
        """Adds IEP skill"""

    def removeSkill(section, skill):
        """Removes IEP skill"""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for groups
"""
import unittest
import doctest

from zope.interface.verify import verifyObject
from zope.app.testing import setup
from zope.component import provideHandler, provideAdapter
from zope.interface import implements

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.relationship.tests import setUpRelationships
#from schooltool.testing.catalog import setUpIntIds, tearDownIntIds
#from schooltool.testing.catalog import setUpCatalogs, tearDownCatalogs
from schooltool.requirement.interfaces import IScoreSystemContainer
from schooltool.cando.interfaces import (
    ISkill,
    )
from schooltool.cando.skill import (
    Skill, SkillSet,
    )


def doctest_Skill():
    """Tests for Skill.

        >>> skill = Skill(title=u'Unit testing')

        >>> verifyObject(ISkill, skill)
        True

        >>> skill.title
        u'Unit testing'

        >>> skill.external_id
        u''

        >>> print repr(Skill('very '*15+'long description'))
        <Skill u'very very very ve...ery long description'>

    """


def doctest_Skill_Equivalency():
    """Tests for Skill equivalency.

        >>> def print_skills(iterable):
        ...     print sorted(iterable, key=lambda i:(i.title, i.__name__))

    Say we have three skills.

        >>> sc = SkillSet('Carpentry')
        >>> hammer = sc.add(Skill('Hammering'))
        >>> pound = sc.add(Skill('Pounding'))
        >>> whack = sc.add(Skill('Whacking'))

        >>> print_skills(hammer.equivalent)
        []

    And we decide to make hammering equivalent to pounding.

        >>> hammer.equivalent.add(pound)

        >>> print_skills(hammer.equivalent)
        [<Skill u'Pounding'>]

        >>> print_skills(pound.equivalent)
        [<Skill u'Hammering'>]

    Then, we make whacking equivalent to pounding.

        >>> whack.equivalent.add(pound)

        >>> print_skills(whack.equivalent)
        [<Skill u'Pounding'>]

    Whacking becomes indirectly equivalent to hammering:

        >>> print_skills(whack.findAllEquivalent())
        [<Skill u'Hammering'>, <Skill u'Pounding'>]

    Because of direct dependence to pounding:

        >>> print_skills(whack.equivalent)
        [<Skill u'Pounding'>]

        >>> print_skills(pound.equivalent)
        [<Skill u'Hammering'>, <Skill u'Whacking'>]

    And if we remove the original dependence, indirect
    dependence is removed as well.

        >>> whack.equivalent.remove(pound)

        >>> print_skills(whack.equivalent)
        []

        >>> print_skills(whack.findAllEquivalent())
        []

    """


class AppStub(dict):
    implements(ISchoolToolApplication)


def setUp(test):
    setup.placefulSetUp()
    setup.setUpTraversal()
    setup.setUpAnnotations()
    provideAdapter(lambda n: AppStub(), (None,), ISchoolToolApplication)
    provideAdapter(lambda a: {}, (ISchoolToolApplication,), IScoreSystemContainer)
    # XXX: no int id or catalog usage yet
    #setUpIntIds(test)
    #setUpCatalogs(test)
    setUpRelationships()

def tearDown(test):
    #tearDownCatalogs(test)
    #tearDownIntIds(test)
    setup.placefulTearDown()


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE |
                   doctest.ELLIPSIS |
                   doctest.REPORT_NDIFF)
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                             optionflags=optionflags),
           ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Selenium Functional Testing Utilities for skills.
"""
import os

from schooltool.testing.selenium import SeleniumLayer

dir = os.path.abspath(os.path.dirname(__file__))
filename = os.path.join(dir, 'stesting.zcml')

skill_selenium_layer = SeleniumLayer(filename,
                                      __name__,
                                      'skill_selenium_layer')

def registerSeleniumSetup():
    try:
        import selenium
    except ImportError:
        return
    from schooltool.testing import registry
    import schooltool.testing.selenium

    def importGlobalSkills(browser, filename):
        browser.query.link('School').click()
        browser.query.link('Skills').click()
        browser.query.link('Import Skill Data').click()
        if filename:
            browser.query.name('xlsfile').type(filename)
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'skill.import_xls', importGlobalSkills))

    def addSkillSet(browser, title, label=None):
        browser.query.link('School').click()
        browser.query.link('Skills').click()
        browser.query.link('Skill Sets').click()
        browser.query.link('Skill Set').click()
        browser.query.name('form.widgets.title').type(title)
        if label is not None:
            browser.query.name('form.widgets.label').type(label)
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'skillset.add', addSkillSet))

    def addSkill(browser, skillset, title, label=None, required=True,
                 external_id=None, scoresystem=None):
        browser.query.link('School').click()
        browser.query.link('Skills').click()
        browser.query.link('Skill Sets').click()
        browser.query.link(skillset).click()
        browser.query.link('Skill').click()
        browser.query.name('form.widgets.title').type(title)
        if label is not None:
            browser.query.name('form.widgets.label').type(label)
        if required:
            browser.query.id('form-widgets-required-0').click()
        else:
            browser.query.id('form-widgets-required-1').click()
        if external_id is not None:
            browser.query.name('form.widgets.external_id').type(external_id)
        if scoresystem is not None:
            browser.query.name('form.widgets.scoresystem:list').ui.set_value(scoresystem)
        else:
            browser.query.name('form.widgets.scoresystem:list').ui.set_value('Competency')
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'skill.add', addSkill))

    def addLayer(browser, title):
        browser.query.link('School').click()
        browser.query.link('Skills').click()
        browser.query.link('Layers').click()
        browser.query.link('Layer').click()
        browser.query.name('form.widgets.title').type(title)
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'layer.add', addLayer))

    def addNode(browser, title, label=None):
        browser.query.link('School').click()
        browser.query.link('Skills').click()
        browser.query.link('Search').click()
        browser.query.link('Node').click()
        browser.query.name('form.widgets.title').type(title)
        if label is not None:
            browser.query.name('form.widgets.label').type(label)
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'node.add', addNode))

registerSeleniumSetup()
del registerSeleniumSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from decimal import Decimal

from zope.annotation.interfaces import IAnnotations
import zc.catalog.extentcatalog
from zope.catalog.text import TextIndex
from zope.index.text.interfaces import ISearchableText
from zope.interface import implements, implementer
from zope.component import adapter, adapts
from zope.container.btree import BTreeContainer
from zope.container.interfaces import INameChooser
from zope.security.proxy import removeSecurityProxy

from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.catalog import AttributeCatalog
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.cando import interfaces
from schooltool.relationship import URIObject
from schooltool.relationship import RelationshipSchema, RelationshipProperty
from schooltool.requirement.interfaces import IScoreSystemContainer
from schooltool.requirement.requirement import Requirement
from schooltool.requirement.scoresystem import CustomScoreSystem
from schooltool.requirement.scoresystem import GlobalDiscreteValuesScoreSystem

DEFAULT_SCORESYSTEM_KEY = 'schooltool.cando.defaultscoresystem'


URISkill = URIObject(
    'http://schooltool.org/ns/cando/skill',
    'Skill',
    'A single skill.')

URISkillSet = URIObject(
    'http://schooltool.org/ns/cando/skillset',
    'Skillset',
    'A set of skills.')

URIEquivalent = URIObject(
    'http://schooltool.org/ns/cando/skill/equivalent',
    'Equivalent',
    'All equivalend skills.')

EquivalentSkills = RelationshipSchema(
    URIEquivalent,
    skill=URISkill,
    equivalent=URISkill)


class Skill(Requirement):
    implements(interfaces.ISkill)

    external_id = u''
    label = u''
    description = u''
    required = False
    retired = False
    custom_scoresystem = None

    equivalent = RelationshipProperty(URIEquivalent, URISkill, URISkill)

    def __init__(self, title, required=False, external_id=u'', label=u'',
                 scoresystem=None):
        Requirement.__init__(self, title)
        self.required = required
        self.external_id = external_id
        self.label = label
        self.custom_scoresystem = scoresystem

    @property
    def scoresystem(self):
        if self.custom_scoresystem is not None:
            return self.custom_scoresystem
        return querySkillScoreSystem()

    @scoresystem.setter
    def scoresystem(self, new_scoresystem):
        self.custom_scoresystem = new_scoresystem

    def findAllEquivalent(self):
        """Find indirectly equivalent skills."""
        visited = set()
        result = list()
        open = set(self.equivalent)
        while open:
            skill = open.pop()
            # XXX: proxies!
            if (skill not in visited and
                skill is not self):
                visited.add(skill)
                result.append(skill)
            open.update(set(skill.equivalent).difference(set(visited)))
        return result

    def copy(self):
        return Skill(title=self.title,
                     required=self.required,
                     external_id=self.external_id,
                     label=self.label)

    def __repr__(self):
        desc = unicode(self.title)
        if len(desc) > 40:
            desc = desc[:17]+'...'+desc[-20:]
        return '<Skill %r>' % unicode(desc)


class SkillSetContainer(BTreeContainer):
    """Container of skill sets."""
    implements(interfaces.ISkillSetContainer)


class SkillSet(Requirement):
    implements(interfaces.ISkillSet)

    description = u''
    label = u''
    retired = False

    def __init__(self, title, description=u'', label=u''):
        Requirement.__init__(self, title)
        self.description = description
        self.label = label

    def add(self, skill):
        skill_copy = skill.copy()
        chooser = INameChooser(self)
        name = chooser.chooseName(skill.__name__, skill_copy)
        self[name] = skill_copy
        return skill_copy


class SkillInit(InitBase):

    def __call__(self):
        self.app['schooltool.cando.skillset'] = SkillSetContainer()


class SkillAppStartup(StartUpBase):
    def __call__(self):
        if 'schooltool.cando.skillset' not in self.app:
            self.app['schooltool.cando.skillset'] = SkillSetContainer()


@implementer(interfaces.ISkillSetContainer)
@adapter(ISchoolToolApplication)
def getSkillSetContainer(app):
    return app['schooltool.cando.skillset']


SkillScoreSystem = GlobalDiscreteValuesScoreSystem(
    'SkillScoreSystem',
    u'Competency', u'Skill Competency Score',
    [('4', u'Expert', Decimal(4), Decimal(90)),
     ('3', u'Competent', Decimal(3), Decimal(70)),
     ('2', u'Practicing', Decimal(2), Decimal(50)),
     ('1', u'Beginning', Decimal(1), Decimal(30)),
     ('0', u'Uninformed', Decimal(0), Decimal(0))],
     '4', '3')


class ScoreSystemAppStartup(StartUpBase):
    after = ('schooltool.requirement.scoresystem', )

    def __call__(self):
        ssc = IScoreSystemContainer(self.app)
        if SkillScoreSystem.__name__ in ssc:
            return
        ssc[SkillScoreSystem.__name__] = CustomScoreSystem(
            SkillScoreSystem.title, SkillScoreSystem.description,
            SkillScoreSystem.scores,
            SkillScoreSystem._bestScore, SkillScoreSystem._minPassingScore)


def querySkillScoreSystem():
    """Get default skill score system for evaluations."""
    app = ISchoolToolApplication(None)
    ssc = IScoreSystemContainer(app)
    ss = ssc.get(SkillScoreSystem.__name__, None)
    if ss is not None:
        return ss
    if len(ssc) > 0:
        return ssc.values()[0]
    return None


def getDefaultSkillScoreSystem(person):
    default_ss = querySkillScoreSystem()
    if default_ss is None:
        return None

    default = [default_ss.__name__.encode('punycode')]
    if person is None:
        return default
    ann = IAnnotations(removeSecurityProxy(person))
    if DEFAULT_SCORESYSTEM_KEY not in ann:
        return default
    return ann[DEFAULT_SCORESYSTEM_KEY]


def setDefaultSkillScoreSystem(person, scoresystem):
    if person is None:
        return
    person = removeSecurityProxy(person)
    ann = IAnnotations(person)
    ann[DEFAULT_SCORESYSTEM_KEY] = scoresystem


def is_global_skillset(index, docid, item):
    if (not interfaces.ISkillSet.providedBy(item) or
        not interfaces.ISkillSetContainer.providedBy(item.__parent__) or
        not ISchoolToolApplication.providedBy(item.__parent__.__parent__)):
        return False
    return True


def is_global_skill(index, docid, item):
    if (not interfaces.ISkill.providedBy(item) or
        not interfaces.ISkillSet.providedBy(item.__parent__) or
        not interfaces.ISkillSetContainer.providedBy(item.__parent__.__parent__) or
        not ISchoolToolApplication.providedBy(item.__parent__.__parent__.__parent__)):
        return False
    return True


searchable_common_indexes = {
    'text_ID': 'getSearchableID',
    'text_title': 'getSearchableTitle',
    'text_label': 'getSearchableLabel',
    }


def setSearchableIndexes(catalog):
    catalog['text'] = TextIndex('getSearchableText', ISearchableText, True)
    for index_id, method_name in searchable_common_indexes.items():
        catalog[index_id] = TextIndex(method_name, ISearchableText, True)


class SkillCatalog(AttributeCatalog):

    version = '1.3 - add __name__ to text index'
    interface = interfaces.ISkill
    attributes = ('title', 'external_id', 'label', 'description',
                  'required', 'retired')

    def createCatalog(self):
        return zc.catalog.extentcatalog.Catalog(
            zc.catalog.extentcatalog.FilterExtent(is_global_skill))

    def setIndexes(self, catalog):
        super(SkillCatalog, self).setIndexes(catalog)
        setSearchableIndexes(catalog)


getSkillCatalog = SkillCatalog.get


class SearchableTextMixin(object):

    def getSearchableID(self):
        return self.context.__name__

    def getSearchableTitle(self):
        return self.context.title

    def getSearchableLabel(self):
        return self.context.label or ''


class SearchableTextSkill(SearchableTextMixin):

    adapts(interfaces.ISkill)
    implements(ISearchableText)

    def __init__(self, context):
        self.context = context

    def getSearchableText(self):
        result = [
            self.context.__name__,
            self.context.title,
            self.context.external_id or '',
            self.context.label or '',
            self.context.description or '',
            ]
        return ' '.join(result)


class SkillSetCatalog(AttributeCatalog):

    version = '1.3 - add __name__ to text index'
    interface = interfaces.ISkillSet
    attributes = ('title', 'label', 'description', 'retired')

    def createCatalog(self):
        return zc.catalog.extentcatalog.Catalog(
            zc.catalog.extentcatalog.FilterExtent(is_global_skillset))

    def setIndexes(self, catalog):
        super(SkillSetCatalog, self).setIndexes(catalog)
        setSearchableIndexes(catalog)


getSkillSetCatalog = SkillSetCatalog.get


class SearchableTextSkillSet(SearchableTextMixin):

    adapts(interfaces.ISkillSet)
    implements(ISearchableText)

    def __init__(self, context):
        self.context = context

    def getSearchableText(self):
        result = [
            self.context.__name__,
            self.context.title,
            self.context.label or '',
            self.context.description or '',
            ]
        return ' '.join(result)


# + directly equivalent
# + all equivalent



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for groups
"""
import unittest
import doctest

from zope.interface.verify import verifyObject
from zope.app.testing import setup
from zope.component import provideHandler
from zope.interface import implements, classImplements
from zope.annotation.interfaces import IAttributeAnnotatable

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.relationship.tests import setUpRelationships
from schooltool.cando.interfaces import (
    ILayer, ILayerContained, ILayerContainer,
    INode, INodeContained, INodeContainer,
    IDocument, IDocumentContained, IDocumentContainer,
    )
from schooltool.cando.model import (
    Layer, LayerContainer,
    NodeContainer, Node,
    preventLayerCycles,
    nodeLinkDoesntViolateModel,
    nodeLayerDoesntViolateModel,
    removingLayerDoesntViolateModel,
    NodeSkillSets,
    DocumentContainer, Document,
    )
from schooltool.cando.skill import (
    Skill, SkillSet,
    )


def doctest_Layers():
    """Tests for Layer.

        >>> layers = LayerContainer()

        >>> verifyObject(ILayerContainer, layers)
        True

        >>> layers['craft'] = Layer('Craft')

        >>> verifyObject(ILayerContained, layers['craft'])
        True

        >>> layers['branch'] = Layer('Branch')
        >>> layers['branch'].parents.add(layers['craft'])

        >>> print list(layers['branch'].parents)
        [<Layer 'Craft'>]

        >>> layers['topic'] = Layer('Topic')
        >>> layers['topic'].parents.add(layers['branch'])

        >>> layers['craft'].parents.add(layers['topic'])
        Traceback (most recent call last):
        ...
        CyclicRelationship:
          child <Layer 'Craft'>
          target parent <Layer 'Topic'>
          distant parents <Layer 'Branch'>, <Layer 'Craft'>

    """


def buildTestModel_crafts():
    layers = LayerContainer()
    layers['craft'] = Layer('Craft')
    layers['branch'] = Layer('Branch')
    layers['branch'].parents.add(layers['craft'])
    layers['topic'] = Layer('Topic')
    layers['topic'].parents.add(layers['branch'])

    nodes = NodeContainer()
    nodes['carpentry'] = Node('Carpentry')
    nodes['carpentry'].layers.add(layers['craft'])
    nodes['creative'] = Node('Creative')
    nodes['creative'].layers.add(layers['branch'])
    nodes['creative'].parents.add(nodes['carpentry'])
    nodes['conventional'] = Node('Conventional')
    nodes['conventional'].layers.add(layers['branch'])
    nodes['conventional'].parents.add(nodes['carpentry'])
    nodes['whacking'] = Node('Whacking')
    nodes['whacking'].layers.add(layers['topic'])
    nodes['whacking'].parents.add(nodes['creative'])
    nodes['hammering'] = Node('Hammering')
    nodes['hammering'].layers.add(layers['topic'])
    nodes['hammering'].parents.add(nodes['conventional'])
    nodes['pounding'] = Node('Pounding')
    nodes['pounding'].layers.add(layers['topic'])
    nodes['pounding'].parents.add(nodes['creative'])
    nodes['pounding'].parents.add(nodes['conventional'])
    return layers, nodes


def buildTestModel_documents():
    layers, nodes = buildTestModel_crafts()
    documents = DocumentContainer()
    main = documents['main'] = Document('Main Document')
    main.hierarchy.add(layers['craft'])
    main.hierarchy.add(layers['branch'])
    main.hierarchy.add(layers['topic'])
    return documents


def doctest_Node():
    """Tests for Layer.

        >>> layers, nodes = buildTestModel_crafts()

        >>> verifyObject(INodeContainer, nodes)
        True

        >>> verifyObject(INodeContained, nodes['carpentry'])
        True

        >>> nodes['pounding']
        <Node 'Pounding' <Layer 'Topic'>>

    """


def doctest_Node_parental_consistency():
    """Test node parent relationship constraints.

        >>> layers, nodes = buildTestModel_crafts()

    Nodes can't have cycles.

        >>> nodes['conventional'].parents.add(nodes['pounding'])
        Traceback (most recent call last):
        ...
        CyclicRelationship:
          child <Node 'Conventional' <Layer 'Branch'>>
          target parent <Node 'Pounding' <Layer 'Topic'>>
          distant parents <Node 'Carpentry' <Layer 'Craft'>>,
                          <Node 'Conventional' <Layer 'Branch'>>,
                          <Node 'Creative' <Layer 'Branch'>>

    Let's try linking nodes that would violate the document model.

        >>> nodes['new_age'] = Node('New age')
        >>> nodes['futuristic'] = Node('Futuristic')

        >>> nodes['futuristic'].layers.add(layers['branch'])
        >>> nodes['futuristic'].parents.add(nodes['new_age'])

    Let's add a futuristic(branch) as a child of a whacking(topic).

        >>> nodes['futuristic'].parents.add(nodes['whacking'])
        Traceback (most recent call last):
        ...
        ViolateLayerModel:
          candidate <Node 'Futuristic' <Layer 'Branch'>>
          acceptable parent layers <Layer 'Craft'>
          violates parent <Node 'Whacking' <Layer 'Topic'>>

    Also, we can't add node that has a child ('branch') to a topic.

        >>> nodes['new_age'].parents.add(nodes['whacking'])
        Traceback (most recent call last):
        ...
        ViolateLayerModel:
          candidate <Node 'Futuristic' <Layer 'Branch'>>
          acceptable parent layers <Layer 'Craft'>
          violates parent <Node 'Whacking' <Layer 'Topic'>>

    But we can add new_age(no layer) to carpentry(craft),
    because it's child futuristic(branch) fits the craft->branch model.

        >>> nodes['new_age'].parents.add(nodes['carpentry'])

        >>> print nodes['whacking'].findPaths()
        [(<Node 'Carpentry' <Layer 'Craft'>>,
          <Node 'Creative' <Layer 'Branch'>>,
          <Node 'Whacking' <Layer 'Topic'>>)]

    Let's consider another example.

        >>> nodes['bashing'] = Node('Bashing')
        >>> nodes['bashing'].layers.add(layers['topic'])

    We can add only direct layer parents.

    That is, we can assign bashing(topic) to conventional(branch):

        >>> nodes['bashing'].parents.add(nodes['conventional'])

    But we can't assign bashing(topic) to carpentry(craft)

        >>> nodes['bashing'].parents.add(nodes['carpentry'])
        Traceback (most recent call last):
        ...
        ViolateLayerModel:
          candidate <Node 'Bashing' <Layer 'Topic'>>
          acceptable parent layers <Layer 'Branch'>
          violates parent <Node 'Carpentry' <Layer 'Craft'>>

    """


def doctest_Node_layer_consistency():
    """Test node layer relationship constraints.

        >>> layers, nodes = buildTestModel_crafts()

        >>> nodes['bashing'] = Node('Bashing')

        >>> nodes['bashing'].layers.add(layers['craft'])

        >>> nodes['bashing'].layers.add(layers['branch'])
        Traceback (most recent call last):
        ...
        InvalidLayerLink:
        setting layer <Layer 'Branch'>
        for node <Node 'Bashing' <Layer 'Craft'>>
        offends node <Node 'Bashing' <Layer 'Craft'>>

        >>> nodes['bashing'].layers.remove(layers['craft'])

        >>> nodes['bashing'].parents.add(nodes['creative'])

        >>> nodes['bashing'].layers.add(layers['craft'])
        Traceback (most recent call last):
        ...
        InvalidLayerLink:
        setting layer <Layer 'Craft'>
        for node <Node 'Bashing' >
        offends node <Node 'Creative' <Layer 'Branch'>>

        >>> nodes['bashing'].layers.add(layers['topic'])

        >>> nodes['bashing'].layers.add(layers['craft'])
        Traceback (most recent call last):
        ...
        InvalidLayerLink:
        setting layer <Layer 'Craft'>
        for node <Node 'Bashing' <Layer 'Topic'>>
        offends node <Node 'Bashing' <Layer 'Topic'>>

        >>> nodes['creative'].layers.remove(layers['branch'])
        Traceback (most recent call last):
        ...
        CannotRemoveLayer: setting layer <Layer 'Branch'>
        for node <Node 'Creative' <Layer 'Branch'>>
        breaks parents <Node 'Carpentry' <Layer 'Craft'>>
        and children <Node 'Bashing' <Layer 'Topic'>>,
                     <Node 'Pounding' <Layer 'Topic'>>,
                     <Node 'Whacking' <Layer 'Topic'>>

    """


def doctest_Node_findPaths():
    r"""Tests for node.findPaths

        >>> layers, nodes = buildTestModel_crafts()

    Say, we also have an alternative, industry:

        >>> layers['industry'] = Layer('Industry')
        >>> layers['job'] = Layer('Job')
        >>> layers['job'].parents.add(layers['industry'])
        >>> layers['topic'].parents.add(layers['job'])

        >>> nodes['transportation'] = Node('Transportation')
        >>> nodes['transportation'].layers.add(layers['industry'])
        >>> nodes['ship_builder'] = Node('Ship builder')
        >>> nodes['ship_builder'].layers.add(layers['job'])
        >>> nodes['ship_builder'].parents.add(nodes['transportation'])

        >>> nodes['construction'] = Node('Construction')
        >>> nodes['construction'].layers.add(layers['industry'])
        >>> nodes['carpenter'] = Node('Carpenter')
        >>> nodes['carpenter'].layers.add(layers['job'])
        >>> nodes['carpenter'].parents.add(nodes['construction'])

        >>> nodes['hammering'].parents.add(nodes['carpenter'])
        >>> nodes['hammering'].parents.add(nodes['ship_builder'])
        >>> nodes['whacking'].parents.add(nodes['carpenter'])
        >>> nodes['pounding'].parents.add(nodes['ship_builder'])

    findPaths will return a list of paths that lead to our node:

        >>> print nodes['pounding'].findPaths()
        [(<Node 'Carpentry' <Layer 'Craft'>>,
          <Node 'Conventional' <Layer 'Branch'>>,
          <Node 'Pounding' <Layer 'Topic'>>),
         (<Node 'Carpentry' <Layer 'Craft'>>,
          <Node 'Creative' <Layer 'Branch'>>,
          <Node 'Pounding' <Layer 'Topic'>>),
         (<Node 'Transportation' <Layer 'Industry'>>,
          <Node 'Ship builder' <Layer 'Job'>>,
          <Node 'Pounding' <Layer 'Topic'>>)]

    Here's another example:

        >>> print ('\n'+'-'*40+'\n').join(
        ...     [str(p) for p in nodes['hammering'].findPaths()])
        (<Node 'Construction' <Layer 'Industry'>>,
         <Node 'Carpenter' <Layer 'Job'>>,
         <Node 'Hammering' <Layer 'Topic'>>)
        ----------------------------------------
        (<Node 'Carpentry' <Layer 'Craft'>>,
         <Node 'Conventional' <Layer 'Branch'>>,
         <Node 'Hammering' <Layer 'Topic'>>)
        ----------------------------------------
        (<Node 'Transportation' <Layer 'Industry'>>,
         <Node 'Ship builder' <Layer 'Job'>>,
         <Node 'Hammering' <Layer 'Topic'>>)

    A lonely node:

        >>> nodes['something'] = Node('Something')
        >>> nodes['something'].findPaths()
        [(<Node 'Something' >,)]

    """


def doctest_Node_skillsets():
    r"""Tests for node.skillsets.

        >>> layers, nodes = buildTestModel_crafts()

    Nodes can have assigned skillsets.

        >>> sc1 = SkillSet('Whacking with a hammer')
        >>> sc2 = SkillSet('Whacking with a shoe')
        >>> sc3 = SkillSet('Work place ethics')

        >>> nodes['whacking'].skillsets.add(sc1)
        >>> nodes['whacking'].skillsets.add(sc2)
        >>> nodes['whacking'].skillsets.add(sc3)

        >>> print sorted(nodes['whacking'].skillsets, key=lambda s: s.title)
        [SkillSet('Whacking with a hammer'),
         SkillSet('Whacking with a shoe'),
         SkillSet('Work place ethics')]

        >>> print list(NodeSkillSets.query(skillset=sc1))
        [<Node 'Whacking' <Layer 'Topic'>>]

    """


def doctest_Document():
    """Tests for Document.

        >>> documents = buildTestModel_documents()

        >>> verifyObject(IDocumentContainer, documents)
        True

        >>> verifyObject(IDocumentContained, documents['main'])
        True

        >>> documents['main']
        <Document 'Main Document' <Layer 'Craft'>, <Layer 'Branch'>, <Layer 'Topic'>>

    """


def setUpModelConstraints(test=None):
    provideHandler(preventLayerCycles)
    provideHandler(nodeLinkDoesntViolateModel)
    provideHandler(nodeLayerDoesntViolateModel)
    provideHandler(removingLayerDoesntViolateModel)


def setUp(test):
    setup.placefulSetUp()
    setup.setUpTraversal()
    setup.setUpAnnotations()
    setUpRelationships()
    setUpModelConstraints(test)

def tearDown(test):
    setup.placefulTearDown()


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE |
                   doctest.ELLIPSIS |
                   doctest.REPORT_NDIFF)
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                             optionflags=optionflags),
           ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Integration with SchoolTool course"""

from persistent.dict import PersistentDict
import zope.lifecycleevent
from zope.annotation.interfaces import IAnnotations
from zope.event import notify
from zope.interface import implements, implementer
from zope.intid.interfaces import IIntIds
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, adapter, getUtility
from zope.container.contained import containedEvent
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent import ObjectModifiedEvent
from zope.proxy.decorator import SpecificationDecoratorBase
from zope.proxy import getProxiedObject
from zope.security.proxy import removeSecurityProxy

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.relationships import CourseSections
from schooltool.app.relationships import URICourseSections
from schooltool.app.relationships import URISectionOfCourse, URICourse
from schooltool.cando.interfaces import ICourseSkills
from schooltool.cando.interfaces import ICourseSkillSet, ICourseSkill
from schooltool.cando.interfaces import ISectionSkills, ISectionSkillSet
from schooltool.cando.interfaces import ISkillSetContainer, ISkillSet, ISkill
from schooltool.cando.skill import Skill, is_global_skillset
from schooltool.course.interfaces import ISection, ICourse
from schooltool.course.interfaces import ICourseContainer
from schooltool.gradebook.activity import Worksheets, GenericWorksheet
from schooltool.requirement.requirement import Requirement
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber

from schooltool.cando import CanDoMessage as _


COURSE_SKILLS_KEY = 'schooltool.cando.project.courseskills'
SECTION_SKILLS_KEY = 'schooltool.cando.project.sectionskills'


class CourseSkills(Requirement):
    implements(ICourseSkills)


class ReadOnlyContainer(KeyError):
    pass


class SectionSkillSet(GenericWorksheet):
    implements(ISectionSkillSet)

    skillset = None

    def __init__(self, skillset):
        self.skillset = skillset
        super(SectionSkillSet, self).__init__(skillset.title)

    @property
    def deployed(self):
        return False

    @property
    def title(self):
        return self.skillset.title

    @title.setter
    def title(self, value):
        pass

    @property
    def description(self):
        course_skillset = self.skillset
        return course_skillset.skillset.description

    @property
    def label(self):
        course_skillset = self.skillset
        return course_skillset.skillset.label

    def all_keys(self):
        return super(SectionSkillSet, self).keys()

    def keys(self):
        return [key for key in self.all_keys()
                if not self[key].retired]

    def __contains__(self, key):
        return key in self.all_keys()


class CourseSkillSet(GenericWorksheet):
    implements(ICourseSkillSet)

    required = None
    retired = None

    def __init__(self, skillset):
        super(CourseSkillSet, self).__init__(skillset.title)
        self.required = PersistentDict()
        self.retired = PersistentDict()

    @Lazy
    def skillset(self):
        if self.__name__ is None:
            return None
        app = ISchoolToolApplication(None)
        ssc = ISkillSetContainer(app)
        return ssc.get(self.__name__)

    def all_keys(self):
        return list(self.skillset.keys())

    def keys(self):
        skillset = self.skillset
        return [key for key in skillset.keys()
                if not self.retired.get(key)]

    def __getitem__(self, key):
        skillset = self.skillset
        skill = skillset[key]
        cs = CourseSkill(skill)
        cs.__parent__ = self
        return cs

    def __setitem__(self, key, newobject):
        raise ReadOnlyContainer(key)

    def __delitem__(self, key):
        raise ReadOnlyContainer(key)


class CourseSkill(SpecificationDecoratorBase):
    """A skill proxy that allows overriding of required/retired attributes."""
    implements(ICourseSkill)

    __slots__ = ('__parent__', )

    @property
    def required(self):
        if self.__name__ not in self.__parent__.required:
            unproxied = getProxiedObject(self)
            return unproxied.required
        return self.__parent__.required[self.__name__]

    @required.setter
    def required(self, value):
        self.__parent__.required[self.__name__] = value

    @property
    def retired(self):
        if self.__name__ not in self.__parent__.retired:
            unproxied = getProxiedObject(self)
            return unproxied.retired
        return self.__parent__.retired[self.__name__]

    @retired.setter
    def retired(self, value):
        self.__parent__.retired[self.__name__] = value


@adapter(ICourse)
@implementer(ICourseSkills)
def getCourseSkills(course):
    annotations = IAnnotations(course)
    try:
        return annotations[COURSE_SKILLS_KEY]
    except KeyError:
        skills = CourseSkills(_('Course Skills'))
        annotations[COURSE_SKILLS_KEY] = skills
        # Sigh, this is not good.
        skills, event = containedEvent(skills, course, 'skills')
        notify(event)
        return skills

getCourseSkills.factory = CourseSkills


@adapter(ICourseSkills)
@implementer(ICourse)
def getCourseSkillsCourse(skills):
    return skills.__parent__


class SectionSkills(Worksheets):
    implements(ISectionSkills)

    annotations_current_worksheet_key = 'schooltool.cando.project.sectionskills'


class SectionSkill(Skill):

    section_intid = None
    source_skillset_name = None
    source_skill_name = None

    @property
    def section(self):
        if self.section_intid is None:
            return None
        int_ids = getUtility(IIntIds)
        section = int_ids.queryObject(self.section_intid)
        return section


@adapter(ISection)
@implementer(ISectionSkills)
def getSectionSkills(section):
    annotations = IAnnotations(section)
    try:
        return annotations[SECTION_SKILLS_KEY]
    except KeyError:
        skills = SectionSkills(_('Section Skills'))
        annotations[SECTION_SKILLS_KEY] = skills
        # Sigh, this is not good.
        skills, event = containedEvent(skills, section, 'skills')
        notify(event)
        return skills

getSectionSkills.factory = SectionSkills


class CourseWorksheetEventSubscriber(ObjectEventAdapterSubscriber):

    @property
    def sections(self):
        skillset = self.object
        course = removeSecurityProxy(ICourse(skillset.__parent__))
        sections = list(CourseSections.query(course=course))
        return sections


class CourseWorksheetRemoved(CourseWorksheetEventSubscriber):
    adapts(IObjectRemovedEvent, ICourseSkillSet)

    def __call__(self):
        skillset = self.object
        for section in self.sections:
            worksheets = ISectionSkills(section)
            if self.object.__name__ in worksheets:
                del worksheets[skillset.__name__]


class ICustomObjectModifiedEvent(IObjectModifiedEvent):

    pass


class CustomObjectModifiedEvent(ObjectModifiedEvent):

    implements(ICustomObjectModifiedEvent)


class GlobalSkillSetUpdateMixin(object):

    def yearsToUpdate(self):
        app = ISchoolToolApplication(None)
        syc = ISchoolYearContainer(app)
        active = syc.getActiveSchoolYear()
        if active is None:
            return list(syc.values())

        idx = removeSecurityProxy(syc.sorted_schoolyears).index(removeSecurityProxy(active))
        years = syc.sorted_schoolyears[idx:]
        return years

    def updateSkillSet(self, skillset):
        years = self.yearsToUpdate()
        for year in years:
            courses = ICourseContainer(year)
            for course in courses.values():
                annotations = IAnnotations(course)
                if COURSE_SKILLS_KEY not in annotations:
                    continue
                course_skills = annotations[COURSE_SKILLS_KEY]
                for course_skillset in course_skills.values():
                    if course_skillset.__name__ == skillset.__name__:
                        notify(CustomObjectModifiedEvent(course_skillset))


class GlobalSkillSetModified(ObjectEventAdapterSubscriber,
                             GlobalSkillSetUpdateMixin):
    adapts(IObjectModifiedEvent, ISkillSet)

    def __call__(self):
        skillset = self.object
        if not is_global_skillset(None, None, skillset):
            return
        self.updateSkillSet(skillset)
        if skillset.retired:
            self.retireSkills(skillset)

    def retireSkills(self, skillset):
        for skill_id in skillset:
            skill = skillset[skill_id]
            if not skill.retired:
                skill.retired = True
                zope.lifecycleevent.modified(skill)


class GlobalSkillModified(ObjectEventAdapterSubscriber,
                          GlobalSkillSetUpdateMixin):
    adapts(IObjectModifiedEvent, ISkill)

    def __call__(self):
        skillset = self.object.__parent__
        if not is_global_skillset(None, None, skillset):
            return
        self.updateSkillSet(skillset)


def updateCourseSkillSet(skillset, section, update_all_attrs=True):
    attrs = ('external_id', 'label', 'description', 'title')
    if update_all_attrs:
        attrs += ('required', 'retired', 'scoresystem')
    int_ids = getUtility(IIntIds)
    worksheets = ISectionSkills(section)
    section_intid = int_ids.getId(section)
    unproxied_skillset = removeSecurityProxy(skillset)
    if skillset.__name__ not in worksheets:
        worksheet = worksheets[skillset.__name__] = SectionSkillSet(unproxied_skillset)
    else:
        worksheet = worksheets[skillset.__name__]
    delete_skills = list(worksheet.all_keys())
    for skill_name in skillset.all_keys():
        skill = skillset[skill_name]
        if skill_name not in worksheet.all_keys():
            target_skill = worksheet[skill_name] = SectionSkill(skill.title)
            target_skill.equivalent.add(removeSecurityProxy(skill))
        else:
            if skill_name in delete_skills:
                delete_skills.remove(skill_name)
            target_skill = worksheet[skill_name]

        for attr in attrs:
            val = getattr(skill, attr, None)
            if getattr(target_skill, attr, None) != val:
                setattr(target_skill, attr, val)
        if target_skill.section_intid != section_intid:
            target_skill.section_intid = section_intid
        if target_skill.source_skill_name != skill.__name__:
            target_skill.source_skill_name = skill.__name__
        if target_skill.source_skillset_name != skill.__parent__.__name__:
            target_skill.source_skillset_name = skill.__parent__.__name__

    available = worksheet.all_keys()
    for skill_name in delete_skills:
        if skill_name in available:
            del worksheet[skill_name]


class CourseSkillSetModified(CourseWorksheetEventSubscriber):

    adapts(IObjectModifiedEvent, ICourseSkillSet)

    def __call__(self):
        skillset = removeSecurityProxy(self.object)
        skillset.title = skillset.skillset.title
        for section in self.sections:
            updateCourseSkillSet(skillset, section)


class CustomCourseSkillSetModified(CourseWorksheetEventSubscriber):

    adapts(ICustomObjectModifiedEvent, ICourseSkillSet)

    def __call__(self):
        skillset = removeSecurityProxy(self.object)
        skillset.title = skillset.skillset.title
        for section in self.sections:
            updateCourseSkillSet(skillset, section, update_all_attrs=False)


class CourseWorksheetAdded(CourseSkillSetModified):
    adapts(IObjectAddedEvent, ICourseSkillSet)


class DeploySkillsToNewSection(ObjectEventAdapterSubscriber):
    adapts(IObjectAddedEvent, ISection)

    def __call__(self):
        section = self.object
        for course in section.courses:
            courseskills = ICourseSkills(course)
            for skillset in courseskills.values():
                updateCourseSkillSet(skillset, section)


def updateSectionSkillsOnCourseChange(event):
    if event.rel_type != URICourseSections:
        return
    section = event[URISectionOfCourse]
    course = event[URICourse]
    courseskills = ICourseSkills(course)
    for skillset in courseskills.values():
        updateCourseSkillSet(skillset, section)


# XXX: maybe course-skillset relationship views
# XXX: helper: linking of all node skillsets to course
# XXX: helper: copying (with resetting required=False) of skillsets
#              (and linking to course)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Skill views.
"""

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts
from zope.container.interfaces import INameChooser
from zope.interface import implements, Interface
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.browser import BrowserView
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL
import z3c.form.field
import z3c.form.form
import z3c.form.button
from z3c.form.browser.text import TextWidget
from z3c.form.widget import FieldWidget
from z3c.form.term import BoolTerms
from z3c.form.interfaces import IRadioWidget
import zc.table.column
import zc.table.interfaces

from schooltool.skin import flourish
from schooltool import table
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.person.interfaces import IPerson
from schooltool.cando.interfaces import ISkillSetContainer
from schooltool.cando.interfaces import ISkillSet, ISkill
from schooltool.cando.interfaces import ISkillRequiredBool
from schooltool.cando.skill import SkillSet, Skill
from schooltool.common.inlinept import InheritTemplate
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.skin.flourish import IFlourishLayer

from schooltool.cando import CanDoMessage as _
from schooltool.cando.skill import getDefaultSkillScoreSystem
from schooltool.cando.skill import setDefaultSkillScoreSystem

class SkillSetContainerView(flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/view/container/table" />
      <h3 tal:condition="python: not len(context)" i18n:domain="schooltool">There are no skill sets.</h3>
    ''')

    @Lazy
    def container(self):
        return ISkillSetContainer(ISchoolToolApplication(None))


class SkillSetTable(table.ajax.Table):

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})

    def columns(self):
        default = table.ajax.Table.columns(self)
        skills = zc.table.column.GetterColumn(
            name='skills',
            title=_(u'Skills'),
            getter=lambda i, f: str(len(i)))
        label = table.column.LocaleAwareGetterColumn(
            name='label',
            title=_(u'Label'),
            getter=lambda i, f: i.label or '',
            subsort=True)
        return [label] + default + [skills]

    def sortOn(self):
        return (('label', False), ('title', False))


class SkillSetTableFilter(table.ajax.TableFilter, table.table.FilterWidget):

    title = _("Title, description or label")

    def filter(self, results):
        if self.ignoreRequest:
            return results
        if 'SEARCH' in self.request:
            searchstr = self.request['SEARCH'].lower()
            results = [item for item in results
                       if searchstr in item.title.lower() or
                       (item.label and searchstr in item.label.lower()) or
                       (item.description and
                        searchstr in item.description.lower())]
        return results


class SkillSetContainerAbsoluteURLAdapter(BrowserView):
    adapts(ISkillSetContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = absoluteURL(app, self.request)
        return url + '/skills'

    __call__ = __str__


class SkillSetContainerLinks(flourish.page.RefineLinksViewlet):
    pass


class SkillSetContainerActionLinks(flourish.page.RefineLinksViewlet):
    pass


class SkillSetAddView(flourish.form.AddForm):

    label = None
    legend = _('Skill set')
    add_next = False

    fields = z3c.form.field.Fields(ISkillSet)
    fields = fields.select('title', 'description', 'label')

    def updateActions(self):
        super(SkillSetAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['submitadd'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        if self.add_next:
            return url + '/add.html'
        return url

    def create(self, data):
        skillset = SkillSet(data['title'])
        z3c.form.form.applyChanges(self, skillset, data)
        self._skillset = skillset
        return skillset

    def add(self, skillset):
        chooser = INameChooser(self.context)
        name = unicode(skillset.title).encode('punycode')
        name = name[:8]+str(len(self.context)+1)
        name = chooser.chooseName(name, skillset)
        self.context[name] = skillset
        return skillset

    @z3c.form.button.buttonAndHandler(_('Submit'), name='add')
    def handleSubmit(self, action):
        super(SkillSetAddView, self).handleAdd.func(self, action)

    @z3c.form.button.buttonAndHandler(_('Submit and add'), name='submitadd')
    def handleSubmitAndAdd(self, action):
        super(SkillSetAddView, self).handleAdd.func(self, action)
        if self._finishedAdd:
            self.add_next = True

    @z3c.form.button.buttonAndHandler(_('Cancel'))
    def handleCancel(self, action):
        super(SkillSetAddView, self).handleCancel.func(self, action)


# XXX: after adding skillset, redirect to it's edit view.

class SkillSetView(flourish.form.DisplayForm):
    template = InheritTemplate(flourish.page.Page.template)
    fields = z3c.form.field.Fields(ISkillSet)
    fields = fields.select('description', 'label', 'retired')


class SkillSetEditView(flourish.form.Form, z3c.form.form.EditForm):
    fields = z3c.form.field.Fields(ISkillSet)
    fields = fields.select('title', 'description', 'label', 'retired')

    legend = _('Skill set')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(SkillSetEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @z3c.form.button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def updateActions(self):
        super(SkillSetEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class SkillSetSkillTable(table.ajax.Table):

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})

    def sortOn(self):
        return (('label', False), ('title', False))

    def columns(self):
        default = table.ajax.Table.columns(self)
        label = table.column.LocaleAwareGetterColumn(
            name='label',
            title=_(u'Label'),
            getter=lambda i, f: i.label or '',
            subsort=True)
        return [label] + default


class SkillSetLinks(flourish.page.RefineLinksViewlet):
    pass


class SkillAddView(flourish.form.AddForm):

    label = None
    legend = _('Skill')
    add_next = False

    fields = z3c.form.field.Fields(ISkill)
    fields = fields.select('title', 'scoresystem', 'description', 'label',
                           'required', 'external_id')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='add')
    def handleSubmit(self, action):
        super(SkillAddView, self).handleAdd.func(self, action)

    @z3c.form.button.buttonAndHandler(_('Submit and add'), name='submitadd')
    def handleSubmitAndAdd(self, action):
        super(SkillAddView, self).handleAdd.func(self, action)
        if self._finishedAdd:
            self.add_next = True

    @z3c.form.button.buttonAndHandler(_('Cancel'))
    def handleCancel(self, action):
        super(SkillAddView, self).handleCancel.func(self, action)

    def updateActions(self):
        super(SkillAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['submitadd'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def updateWidgets(self):
        super(SkillAddView, self).updateWidgets()

        scoresystem = self.widgets['scoresystem'].value
        if scoresystem:
            return
        person = IPerson(self.request.principal, None)
        default = getDefaultSkillScoreSystem(person)
        if default is not None:
            self.widgets['scoresystem'].value = default

    def nextURL(self):
        url = absoluteURL(self.context, self.request)
        if self.add_next:
            return url + '/add.html'
        return url

    def create(self, data):
        if not data['label']:
            data['label'] = u'%02d' % (len(self.context) + 1)
        skill = Skill(data['title'])
        z3c.form.form.applyChanges(self, skill, data)
        self._skill = skill
        return skill

    def add(self, skill):
        chooser = INameChooser(self.context)
        if skill.external_id:
            name = skill.external_id
        else:
            name = unicode(skill.title).encode('punycode')
            name = name[:8]+str(len(self.context)+1)
        name = chooser.chooseName(name, skill)
        self.context[name] = skill
        scoresystem = self.request.get(self.widgets['scoresystem'].name, '')
        if scoresystem:
            person = IPerson(self.request.principal, None)
            setDefaultSkillScoreSystem(person, scoresystem)
        return skill


class SkillView(flourish.form.DisplayForm):

    template = InheritTemplate(flourish.page.Page.template)

    label = None
    legend = _('Skill')

    fields = z3c.form.field.Fields(ISkill)
    fields = fields.select('scoresystem', 'description', 'label', 'required',
                           'retired', 'external_id')

    @property
    def title(self):
        return self.context.__parent__.title

    @property
    def can_edit(self):
        return flourish.canEdit(self.context)

    @property
    def edit_url(self):
        return absoluteURL(self.context, self.request) + '/edit.html'

    @property
    def done_url(self):
        return absoluteURL(self.context.__parent__, self.request)


class SkillEditView(flourish.form.Form, z3c.form.form.EditForm):
    fields = z3c.form.field.Fields(ISkill)
    fields = fields.select('title', 'description', 'label',
                           'required', 'external_id')

    legend = _('Skill')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(SkillEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @z3c.form.button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def updateActions(self):
        super(SkillEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class LabelTextLineWidget(TextWidget):

    def update(self):
        super(LabelTextLineWidget, self).update()
        self.maxlength = self.field.max_length


def LabelTextLineFieldWidget(field, request):
    return FieldWidget(field, LabelTextLineWidget(request))


class SkillRequiredTerms(BoolTerms):

    adapts(Interface,
           IFlourishLayer,
           Interface,
           ISkillRequiredBool,
           IRadioWidget)

    trueLabel = _('Required')
    falseLabel = _('Optional')



#!/usr/bin/env python
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012, 2013 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool CanDo setup script.
"""

import os
from setuptools import setup, find_packages

if os.path.exists("version.txt"):
    version = open("version.txt").read().strip()
else:
    version = open("version.txt.in").read().strip()

def read(*rnames):
    text = open(os.path.join(os.path.dirname(__file__), *rnames)).read()
    return text

setup(
    name="schooltool.cando",
    description="CanDo plugin for SchoolTool",
    long_description=(
        read('README.txt')
        + '\n\n' +
        read('CHANGES.txt')
        ),
    version=version,
    url='http://www.schooltool.org',
    license="GPL",
    maintainer="SchoolTool Developers",
    maintainer_email="schooltool-developers@lists.launchpad.net",
    platforms=["any"],
    classifiers=["Development Status :: 4 - Beta",
    "Environment :: Web Environment",
    "Intended Audience :: End Users/Desktop",
    "License :: OSI Approved :: GNU General Public License (GPL)",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Zope",
    "Topic :: Education"],
    package_dir={'': 'src'},
    packages=find_packages('src'),
    namespace_packages=["schooltool"],
    install_requires=['schooltool >= 2.6',
                      'schooltool.gradebook >= 2.6',
                      'setuptools',
                      'zope.i18n',
                      'zope.i18nmessageid'],
    extras_require={'test': ['schooltool.devtools>=0.6',
                             'z3c.form>=2.6']},
    include_package_data=True,
    zip_safe=False,
    entry_points="""
        [z3c.autoinclude.plugin]
        target = schooltool
        """,
    )
